"""Collection of Enumerations for Units used in ResSimpy."""
from enum import Enum


# Enum representing the Salinity Units
class SUnits(str, Enum):
    MEQ_ML = 'MEQ/ML'
    PPM = 'PPM'


class TemperatureUnits(str, Enum):
    FAHR = 'FAHR'
    CELSIUS = 'CELSIUS'
    RANKINE = 'RANKINE'
    KELVIN = 'KELVIN'


class UnitSystem(str, Enum):
    ENGLISH = 'ENGLISH'  # defined as 'ENGLISH' in Nexus, 'FIELD' in Eclipse
    LAB = 'LAB'  # defined as 'LAB' in Nexus, 'LAB' in Eclipse
    METRIC = 'METRIC'  # defined as 'METRIC' in Nexus, undefined in Eclipse
    METKGCM2 = 'METKG/CM2'  # defined as 'METKG/CM2' in Nexus, undefined in Eclipse
    METBAR = 'METBAR'  # defined as 'METBAR' in Nexus, 'METRIC' in Eclipse
    METRIC_ATM = 'PVT-M'  # undefined in Nexus, PVT-M in Eclipse



HYD_RATE_KEYWORDS = ['QOIL', 'QLIQ', 'QGAS', 'QWAT', 'QWGAS']
HYD_GAS_RATIO_KEYWORDS = ['GOR', 'GLR', 'OGR', 'GWR', 'MMW']
HYD_WATER_RATIO_KEYWORDS = ['WCUT', 'WGR', 'OCUT', 'WWGR']
HYD_PRESSURE_KEYWORDS = ['PIN', 'POUT', 'THP']
HYD_ARRAY_KEYWORDS = HYD_RATE_KEYWORDS + HYD_GAS_RATIO_KEYWORDS + HYD_WATER_RATIO_KEYWORDS + HYD_GAS_RATIO_KEYWORDS + \
    HYD_PRESSURE_KEYWORDS
HYD_ALQ_KEYWORD = ['ALQ']
HYD_ALQ_OPTIONS = ['GASRATE', 'NOTGLIFT']

HYD_CORRECTION_KEYWORD = ['DATGRAD']
HYD_CORRECTION_OPTIONS = ['GRAD', 'OILGRAD', 'GASGRAD', 'WATERGRAD']

HYD_KEYWORDS_VALUE_FLOAT = ['LENGTH', 'DATUM', 'DZW']
HYD_SINGLE_KEYWORDS = ['NOCHK']

HYD_TABLE_HEADER_COLS = ['IGOR', 'IWCUT', 'IQOIL', 'IALG', 'BHP(ITHP)', 'IGLR', 'IQLIQ', 'IOGR', 'IWGR', 'IQGAS',
                         'IGWR', 'IOCUT', 'IQWAT', 'IMMW', 'IWWGR', 'IQWGAS']

HYD_LIMITS_KEYWORD = ['LIMITS']

HYD_WATINJ_KEYWORD = ['WATINJ']
HYD_WATINJ_KEYWORDS_VALUE_FLOAT = ['GRAD', 'VISC', 'DIAM', 'LENGTH', 'DZ', 'ROUGHNESS']

HYD_KEYWORDS = HYD_ARRAY_KEYWORDS + HYD_ALQ_KEYWORD + HYD_CORRECTION_KEYWORD + HYD_KEYWORDS_VALUE_FLOAT + \
    HYD_SINGLE_KEYWORDS + HYD_LIMITS_KEYWORD + HYD_WATINJ_KEYWORD



# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

EOS_SEPARATOR_KEYWORDS = ['STAGE', 'TEMP', 'PRES', 'METHOD', 'IDL1', 'IDL2', 'IDV1', 'IDV2', 'FDL1', 'FDV1', 'IGP',
                          'WATERMETHOD']
GAS_PLANT_KEYWORDS = ['KEYCPTLIST', 'GPTNUM', 'PRES_STD', 'TEMP_STD', 'EOSMETHOD', 'WATERMETHOD', 'RECFAC_TABLE'
                      'ENDRECFAC_TABLE', 'KEYCPTMF']

BLACKOIL_SEPARATOR_KEYWORDS = ['BOSEP', 'MWOIL', 'MWGAS', 'ZOIL', 'STAGE', 'KVOIL', 'KVGAS', 'IDL1', 'IDL2', 'IDV1',
                               'IDV2', 'FDL1', 'FDV1']

SEPARATOR_KEYS_INT = ['WATERMETHOD', 'EOSMETHOD', 'GPTNUM']

SEPARATOR_KEYS_FLOAT = ['MWOIL', 'MWGAS', 'ZOIL', 'PRES_STD', 'TEMP_STD']

SEPARATOR_KEYWORDS = EOS_SEPARATOR_KEYWORDS + GAS_PLANT_KEYWORDS + BLACKOIL_SEPARATOR_KEYWORDS



"""Key functions for reading in nexus constraints and populating objects with the data."""
from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from ResSimpy.Nexus.DataModels.Network.NexusConstraint import NexusConstraint
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.nexus_file_operations import get_next_value, correct_datatypes
from ResSimpy.Utils.invert_nexus_map import nexus_keyword_to_attribute_name
import fnmatch
if TYPE_CHECKING:
    from ResSimpy.File import File


def load_inline_constraints(file_as_list: list[str], constraint: type[NexusConstraint], current_date: Optional[str],
                            unit_system: UnitSystem, property_map: dict[str, tuple[str, type]],
                            existing_constraints: dict[str, list[NexusConstraint]], nexus_file: File,
                            start_line_index: int, network_names: Optional[list[str]] = None) -> None:
    """Loads table of constraints with the wellname/node first and the constraints following inline
        uses previous set of constraints as still applied to the well.

    Args:
    ----
        file_as_list (list[str]): file represented as a list of strings
        constraint (NexusConstraint): object to store the attributes extracted from each row.
        current_date (str): the current date in the table
        unit_system (UnitSystem): Unit system enum
        property_map (dict[str, tuple[str, type]]): Mapping of nexus keywords to attributes
        existing_constraints (dict[str, NexusConstraint]): all existing constraints from previous lines of the \
            surface file
        network_names (Optional[list[str]]): list of names for all nodes, wells and connections in a nexus network.
            Used in deriving constraints from wildcards. Defaults to None

    Returns:
    -------
        dict[UUID, int]: dictionary of object locations derived from inline table.
    """

    for index, line in enumerate(file_as_list):
        properties_dict: dict[str, str | float | UnitSystem | None] = {'date': current_date, 'unit_system': unit_system}
        # first value in the line has to be the node/wellname
        name = get_next_value(0, [line])
        nones_overwrite = False
        constraint_names_to_add: list[str] = []
        if name is None:
            continue
        properties_dict['name'] = name
        if '*' in name:
            if network_names is None:
                raise ValueError('No existing nodes found to add wildcards to')
            else:
                # filter names that match the pattern
                constraint_names_to_add = fnmatch.filter(network_names, name)
        else:
            constraint_names_to_add.append(name)

        trimmed_line = line.replace(name, "", 1)
        next_value = get_next_value(0, [trimmed_line])
        # loop through the line for each set of constraints
        while next_value is not None:
            token_value = next_value.upper()
            if token_value in ['CLEAR', 'CLEARP', 'CLEARQ', 'CLEARLIMIT', 'CLEARALQ']:
                properties_dict[nexus_keyword_to_attribute_name(constraint.get_keyword_mapping(), token_value)] = True
                nones_overwrite = True
                # break out of the while loop as the next value will not be there
                break
            elif token_value == 'ACTIVATE' or token_value == 'DEACTIVATE':
                properties_dict.update({'active_node': token_value == 'ACTIVATE'})
                trimmed_line = trimmed_line.replace(next_value, "", 1)
                next_value = get_next_value(0, [trimmed_line])
                if next_value is None:
                    break
                token_value = next_value.upper()

            trimmed_line = trimmed_line.replace(next_value, "", 1)
            # extract the attribute name for the given nexus constraint token
            if property_map.get(token_value, None) is None:
                # if the next token found along isn't a valid property then move to the next line.
                break

            attribute = property_map[token_value][0]
            next_value = get_next_value(0, [trimmed_line])
            if next_value is None:
                raise ValueError(f'No value found after {token_value} in {line}')
            elif next_value == 'MULT':
                try:
                    attribute = property_map[token_value + '_MULT'][0]
                except AttributeError:
                    raise AttributeError(f'Unexpected MULT keyword following {token_value}')
                properties_dict[attribute] = True

            else:
                properties_dict[attribute] = correct_datatypes(next_value, float)
            trimmed_line = trimmed_line.replace(next_value, "", 1)
            next_value = get_next_value(0, [trimmed_line])

        # first check if there are any existing constraints created for the well this timestep
        for name_of_node in constraint_names_to_add:
            properties_dict['name'] = name_of_node
            well_constraints = existing_constraints.get(name_of_node, None)
            if well_constraints is not None:
                latest_constraint = well_constraints[-1]
                if latest_constraint.date == current_date:
                    latest_constraint.update(properties_dict, nones_overwrite)
                    nexus_file.add_object_locations(latest_constraint.id, [index + start_line_index])
                else:
                    # otherwise take a copy of the previous constraint and add the additional properties
                    new_constraint = constraint(properties_dict)
                    well_constraints.append(new_constraint)
                    nexus_file.add_object_locations(new_constraint.id, [index + start_line_index])
            else:
                nexus_constraint = constraint(properties_dict)

                existing_constraints[name_of_node] = [nexus_constraint]
                nexus_file.add_object_locations(nexus_constraint.id, [index + start_line_index])


def __clear_constraints(token_value, constraint) -> dict[str, None]:
    """Replicates behaviour of the clear keyword in nexus constraints by creating a dictionary filled with
    Nones for the relevant parameters.
    """
    match token_value:
        case 'CLEAR':
            constraint_clearing_dict = constraint.get_rate_constraints_map()
            constraint_clearing_dict.update(constraint.get_pressure_constraints_map())
            constraint_clearing_dict.update(constraint.get_limit_constraints_map())
        case 'CLEARQ':
            constraint_clearing_dict = constraint.get_rate_constraints_map()
        case 'CLEARLIMIT':
            constraint_clearing_dict = constraint.get_limit_constraints_map()
        case 'CLEARP':
            constraint_clearing_dict = constraint.get_pressure_constraints_map()
        case 'CLEARALQ':
            constraint_clearing_dict = constraint.get_alq_constraints_map()
        case _:
            constraint_clearing_dict = {}
    return {x[0]: None for x in constraint_clearing_dict.values()}



"""A library to help read and manipulate input decks for reservoir simulators.

.. autosummary::
    :toctree: _autosummary
    :caption: API Reference
    :template: custom-module-template.rst
    :recursive:

    Nexus
    Utils
    Enums
"""

__version__ = "0.0.0"  # Set at build time



"""Class for handling the mapping between unit systems and the units used for that unit system."""
from abc import ABC

from ResSimpy.Enums.UnitsEnum import UnitSystem

"""
Variable	English Oilfield Units	Metric Units (kPa)	Metric Units (kg/cm2)	Metric Units (bars)	Lab Units
Acoustic Impedance	ft/sec·(g/cm3)	m/sec·(kg/m3)	m/sec·(kg/m3)	m/sec·(kg/m3)	cm/sec·(g/cm3)
Acoustic Wave Velocity	ft/sec	m/sec	m/sec	m/sec	m/sec
Angle	degrees	degrees	degrees	degrees	degrees
Area	ft2	m2	m2	m2	cm2
Bulk Modulus	psia	kPa	kg/cm2	bars	psia
Compressibility	psi-1	kPa-1	(kg/cm2)-1	(bars)-1	psi-1
Critical Pressure	psia	kPa	(kg/cm2)	bars	psia
Critical Temperature	degrees F	degrees C	degrees C	degrees C	degrees C
Critical Volume ft3/lbmole  m3/kgmole   m3/kgmole   m3/kgmole   cc/gmole
Density	lb/ft3	kg/m3	kg/m3	kg/m3	gm/cc
Formation Volume Factor (Oil)	rb/stb	m3/stm3	m3/stm3	m3/stm3	cc/stcc
Formation Volume Factor (Gas)	rb/mscf	m3/stm3	m3/stm3	m3/stm3	cc/stcc
Gas-Liquid Ratio (Gas-Oil Ratio)	mscf/stb	stm3/stm3	stm3/stm3	stm3/stm3	stcc/stcc
Gravity Gradient	psi/ft	kPa/m	kg/cm2/m	bars/m	psi/cm
Heat Capacity   BTU/lb-F    kJ/kg-K kJ/kg-K kJ/kg-K J/g-K
Length	ft	m	m	m	cm
Liquid-Gas Ratio (Oil-Gas Ratio)    stb/mscf    stm3/stm3	stm3/stm3	stm3/stm3	stcc/stcc
Moles	lb-moles	kg-moles	kg-moles	kg-moles	gm-moles
Molar Density	lb-moles/ft3	kg-moles/m3	kg-moles/m3	kg-moles/m3	gm-moles/cm3
Molar Rates	lb-moles/day	kg-moles/day	kg-moles/day	kg-moles/day	gm-moles/hour
Permeability	md	md	md	md	md
Pressure	psia	kPa	kg/cm2	bars	psia
Reservoir Rates	rb/day	m3/day	m3/day	m3/day	cc/hour
Reservoir Volumes	mrb	k m3	k m3	k m3	k cc
Saturation Fraction	fraction	fraction	fraction	fraction	fraction
Solution Gas-Oil Ratio	mscf/stb	stm3/stm3	stm3/stm3	stm3/stm3	stcc/stcc
Solution Oil-Gas Ratio	stb/mscf    stm3/stm3	stm3/stm3	stm3/stm3	stcc/stcc
Surface Rates (Liquid)	stb/day	stm3/day	stm3/day	stm3/day	stcc/hour
Surface Rates (Gas)	mscf/day	stm3/day	stm3/day	stm3/day	stcc/hour
Surface Volumes (Gas)	mmscf	k stm3	k stm3	k stm3	k stcc
Surface Volumes (Liquid)	mstb	k stm3	k stm3	k stm3	k stcc
Temperature	degrees F	degrees C	degrees C	degrees C	degrees C
Time	days	days	days	days	hours
Tracer Concentrations	fraction	fraction	fraction	fraction	fraction
Transmissibility	ft3·cp/day/psi	m3·cp/day/kPa	m3·cp/day/kg/cm2	m3·cp/day/bars	cc·cp/hour/psi
Viscosity	cp	cp	cp	cp	cp
Volume	ft3	m3	m3  m3	cc
"""


class UnitDimension(ABC):
    """Class for handling the mapping between unit systems and the units used for that unit system."""
    english: str = ''
    metric: str = ''
    lab: str = ''
    metkgcm2: str = ''
    metbar: str = ''
    metric_atm: str = ''

    def unit_system_enum_to_variable(self, unit_system: UnitSystem) -> str:
        """Returns the unit variable for the given unit system."""
        match unit_system:
            case UnitSystem.ENGLISH:
                return self.english
            case UnitSystem.LAB:
                return self.lab
            case UnitSystem.METRIC:
                return self.metric
            case UnitSystem.METKGCM2:
                return self.metkgcm2
            case UnitSystem.METBAR:
                return self.metbar
            case UnitSystem.METRIC_ATM:
                return self.metric_atm
            case _:
                raise ValueError(f'Unit system {unit_system} not recognised.')


class Area(UnitDimension):
    """Units for area."""
    english = 'ft2'
    metric = 'm2'
    lab = 'cm2'
    metkgcm2 = 'm2'
    metbar = 'm2'
    metric_atm = 'm2'


class AcousticImpedance(UnitDimension):
    """Units for acoustic impedance."""
    english = 'ft/sec(g/cm3)'
    metric = 'm/sec(kg/m3)'
    metkgcm2 = 'm/sec(kg/m3)'
    metbar = 'm/sec(kg/m3)'
    lab = 'cm/sec(g/cm3)'
    metric_atm = 'm/sec(kg/m3)'


class AcousticWaveVelocity(UnitDimension):
    """Units for acoustic wave velocity."""
    english = 'ft/sec'
    metric = 'm/sec'
    metkgcm2 = 'm/sec'
    metbar = 'm/sec'
    lab = 'm/sec'


class Angle(UnitDimension):
    """Units for angle."""
    english = 'degrees'
    metric = 'degrees'
    lab = 'degrees'
    metkgcm2 = 'degrees'
    metbar = 'degrees'
    metric_atm = 'degrees'


class BulkModulus(UnitDimension):
    """Units for bulk modulus."""
    english = 'psia'
    metric = 'kPa'
    metkgcm2 = 'kg/cm2'
    metbar = 'bars'
    lab = 'psia'
    metric_atm = 'atm'


class Compressibility(UnitDimension):
    """Units for compressibility."""
    english = 'psi-1'
    metric = 'kPa-1'
    metkgcm2 = '(kg/cm2)-1'
    metbar = '(bars)-1'
    lab = 'psi-1'
    metric_atm = '(atm)-1'


class CriticalPressure(UnitDimension):
    """Units for critical pressure."""
    english = 'psia'
    metric = 'kPa'
    metkgcm2 = 'kg/cm2'
    metbar = 'bars'
    lab = 'psia'
    metric_atm = 'atm'


class CriticalTemperature(UnitDimension):
    """Units for critical temperature."""
    english = 'degrees F'
    metric = 'degrees C'
    metkgcm2 = 'degrees C'
    metbar = 'degrees C'
    lab = 'degrees C'


class CriticalVolume(UnitDimension):
    """Units for critical volume."""
    english = 'ft3/lbmole'
    metric = 'm3/kgmole'
    metkgcm2 = 'm3/kgmole'
    metbar = 'm3/kgmole'
    lab = 'cc/gmole'
    metric_atm = 'm3/kgmole'


class Density(UnitDimension):
    """Units for density."""
    english = 'lb/ft3'
    metric = 'kg/m3'
    metkgcm2 = 'kg/m3'
    metbar = 'kg/m3'
    lab = 'gm/cc'
    metric_atm = 'kg/m3'


class FormationVolumeFactorLiquid(UnitDimension):
    """Units for formation volume factor oil."""
    english = 'rb/stb'
    metric = 'm3/stm3'
    metkgcm2 = 'm3/stm3'
    metbar = 'm3/stm3'
    lab = 'cc/stcc'
    metric_atm = 'm3/stm3'


class FormationVolumeFactorGas(UnitDimension):
    """Units for formation volume factor gas."""
    english = 'rb/mscf'
    metric = 'm3/stm3'
    metkgcm2 = 'm3/stm3'
    metbar = 'm3/stm3'
    lab = 'cc/stcc'
    metric_atm = 'm3/stm3'


class GasLiquidRatio(UnitDimension):
    """Units for gas liquid ratio."""
    english = 'mscf/stb'
    metric = 'stm3/stm3'
    metkgcm2 = 'stm3/stm3'
    metbar = 'stm3/stm3'
    lab = 'stcc/stcc'
    metric_atm = 'stm3/stm3'


class GravityGradient(UnitDimension):
    """Units for gravity gradient."""
    english = 'psi/ft'
    metric = 'kPa/m'
    metkgcm2 = 'kg/cm2/m'
    metbar = 'bars/m'
    lab = 'psi/cm'
    metric_atm = 'atm/m'


class Length(UnitDimension):
    """Units for length."""
    english = 'ft'
    metric = 'm'
    lab = 'cm'
    metkgcm2 = 'm'
    metbar = 'm'
    metric_atm = 'm'


class Moles(UnitDimension):
    """Units for moles."""
    english = 'lb-moles'
    metric = 'kg-moles'
    metkgcm2 = 'kg-moles'
    metbar = 'kg-moles'
    lab = 'gm-moles'
    metric_atm = 'kg-moles'


class MolarDensity(UnitDimension):
    """Units for molar density."""
    english = 'lb-moles/ft3'
    metric = 'kg-moles/m3'
    metkgcm2 = 'kg-moles/m3'
    metbar = 'kg-moles/m3'
    lab = 'gm-moles/cm3'
    metric_atm = 'kg-moles/m3'


class MolarRates(UnitDimension):
    """Units for molar rates."""
    english = 'lb-moles/day'
    metric = 'kg-moles/day'
    metkgcm2 = 'kg-moles/day'
    metbar = 'kg-moles/day'
    lab = 'gm-moles/hour'
    metric_atm = 'kg-moles/day'


class Permeability(UnitDimension):
    """Units for permeability."""
    english = 'md'
    metric = 'md'
    metkgcm2 = 'md'
    metbar = 'md'
    lab = 'md'
    metric_atm = 'md'


class Pressure(UnitDimension):
    """Units for pressure."""
    english = 'psia'
    metric = 'kPa'
    metkgcm2 = 'kg/cm2'
    metbar = 'bars'
    lab = 'psia'
    metric_atm = 'atm'


class DeltaPressure(UnitDimension):
    """Units for pressure."""
    english = 'psi'
    metric = 'kPa'
    metkgcm2 = 'kg/cm2'
    metbar = 'bars'
    lab = 'psi'
    metric_atm = 'atm'


class ReservoirRates(UnitDimension):
    """Units for reservoir rates."""
    english = 'rb/day'
    metric = 'm3/day'
    metkgcm2 = 'm3/day'
    metbar = 'm3/day'
    lab = 'cc/hour'
    metric_atm = 'm3/day'


class ReservoirVolumeThousand(UnitDimension):
    """Units for reservoir volumes, in thousands."""
    english = 'mrb'
    metric = 'k m3'
    metkgcm2 = 'k m3'
    metbar = 'k m3'
    lab = 'k cc'
    metric_atm = 'k m3'


class SaturationFraction(UnitDimension):
    """Units for saturation fraction."""
    english = 'fraction'
    metric = 'fraction'
    lab = 'fraction'
    metkgcm2 = 'fraction'
    metbar = 'fraction'
    metric_atm = 'fraction'


class SolutionGasOilRatio(UnitDimension):
    """Units for solution gas oil ratio."""
    english = 'mscf/stb'
    metric = 'stm3/stm3'
    metkgcm2 = 'stm3/stm3'
    metbar = 'stm3/stm3'
    lab = 'stcc/stcc'
    metric_atm = 'stm3/stm3'


class SolutionOilGasRatio(UnitDimension):
    """Units for solution oil gas ratio."""
    english = 'stb/mscf'
    metric = 'stm3/stm3'
    metkgcm2 = 'stm3/stm3'
    metbar = 'stm3/stm3'
    lab = 'stcc/stcc'
    metric_atm = 'stm3/stm3'


class SurfaceRatesLiquid(UnitDimension):
    """Units for surface rates liquid."""
    english = 'stb/day'
    metric = 'stm3/day'
    metkgcm2 = 'stm3/day'
    metbar = 'stm3/day'
    lab = 'stcc/hour'
    metric_atm = 'stm3/day'


class SurfaceRatesGas(UnitDimension):
    """Units for surface rates gas."""
    english = 'mscf/day'
    metric = 'stm3/day'
    metkgcm2 = 'stm3/day'
    metbar = 'stm3/day'
    lab = 'stcc/hour'
    metric_atm = 'stm3/day'


class SurfaceVolumesGas(UnitDimension):
    """Units for surface volumes gas."""
    english = 'mmscf'
    metric = 'k stm3'
    metkgcm2 = 'k stm3'
    metbar = 'k stm3'
    lab = 'k stcc'
    metric_atm = 'k stm3'


class SurfaceVolumesLiquid(UnitDimension):
    """Units for surface volumes liquid."""
    english = 'mstb'
    metric = 'k stm3'
    metkgcm2 = 'k stm3'
    metbar = 'k stm3'
    lab = 'k stcc'
    metric_atm = 'k stm3'


class Temperature(UnitDimension):
    """Units for temperature."""
    english = 'degrees F'
    metric = 'degrees C'
    metkgcm2 = 'degrees C'
    metbar = 'degrees C'
    lab = 'degrees C'
    metric_atm = 'degrees C'


class Time(UnitDimension):
    """Units for time."""
    english = 'days'
    metric = 'days'
    metkgcm2 = 'days'
    metbar = 'days'
    lab = 'hours'
    metric_atm = 'days'


class InverseTime(UnitDimension):
    """Units for 1/time."""
    english = '1/days'
    metric = '1/days'
    metkgcm2 = '1/days'
    metbar = '1/days'
    lab = '1/hours'
    metric_atm = '1/days'


class TracerConcentrations(UnitDimension):
    """Units for tracer concentrations."""
    english = 'fraction'
    metric = 'fraction'
    metkgcm2 = 'fraction'
    metbar = 'fraction'
    lab = 'fraction'
    metric_atm = 'fraction'


class Transmissibility(UnitDimension):
    """Units for transmissibility."""
    english = 'ft3*cp/day/psi'
    metric = 'm3*cp/day/kPa'
    metkgcm2 = 'm3*cp/day/kg/cm2'
    metbar = 'm3*cp/day/bars'
    lab = 'cc*cp/hour/psi'
    metric_atm = 'm3*cp/day/atm'


class Viscosity(UnitDimension):
    """Units for viscosity."""
    english = 'cp'
    metric = 'cp'
    metkgcm2 = 'cp'
    metbar = 'cp'
    lab = 'cp'
    metric_atm = 'cp'


class Volume(UnitDimension):
    """Units for volume."""
    english = 'ft3'
    metric = 'm3'
    metkgcm2 = 'm3'
    metbar = 'm3'
    lab = 'cc'
    metric_atm = 'm3'


class ReservoirVolume(UnitDimension):
    """Units for reservoir volume."""
    english = 'rb'
    metric = 'm3'
    metkgcm2 = 'm3'
    metbar = 'm3'
    lab = 'cc'
    metric_atm = 'm3'


class LiquidGasRatio(UnitDimension):
    """Liquid gas ratio units."""
    english = 'stb/mscf'
    metric = 'stm3/stm3'
    metkgcm2 = 'stm3/stm3'
    metbar = 'stm3/stm3'
    lab = 'stcc/stcc'
    metric_atm = 'stm3/stm3'


class Roughness(UnitDimension):
    """Roughness units."""
    english = 'in'
    metric = 'mm'
    metkgcm2 = 'mm'
    metbar = 'mm'
    lab = 'mm'
    metric_atm = 'mm'


class Diameter(UnitDimension):
    """Diameter units."""
    english = 'in'
    metric = 'cm'
    metkgcm2 = 'cm'
    metbar = 'cm'
    lab = 'cm'
    metric_atm = 'cm'


class HeatTransfer(UnitDimension):
    """Heat transfer units."""
    english = 'BTU/(hr*ft2*F)'
    metric = 'W/(m2*K)'
    metkgcm2 = 'W/(m2*K)'
    metbar = 'W/(m2*K)'
    lab = 'W/(cm2*K)'
    metric_atm = 'W/(m2*K)'


class ThermalConductivity(UnitDimension):
    """Thermal conductivity units."""
    english = 'BTU/(hr*ft*F)'
    metric = 'W/(m*K)'
    metkgcm2 = 'W/(m*K)'
    metbar = 'W/(m*K)'
    lab = 'W/(cm*K)'
    metric_atm = 'W/(m*K)'


class HeatCapacity(UnitDimension):
    """Heat capacity units."""
    english = 'BTU/(lb*F)'
    metric = 'kJ/(kg*K)'
    metkgcm2 = 'kJ/(kg*K)'
    metbar = 'kJ/(kg*K)'
    lab = 'J/(g*K)'
    metric_atm = 'kJ/(kg*K)'


class DiffusionCoefficient(UnitDimension):
    """Diffusion coefficient units."""
    english = 'ft2/day'
    metric = 'm2/day'
    metkgcm2 = 'm2/day'
    metbar = 'm2/day'
    lab = 'cm2/hr'
    metric_atm = 'm2/day'


class Dimensionless(UnitDimension):
    """For attributes which do not have a dimension."""
    english = ''
    metric = ''
    metkgcm2 = ''
    metbar = ''
    lab = ''
    metric_atm = ''


class PermeabilityThickness(UnitDimension):
    """Permeability thickness units."""
    english = 'md*ft'
    metric = 'md*m'
    metkgcm2 = 'md*m'
    metbar = 'md*m'
    lab = 'md*m'
    metric_atm = 'md*m'


class NonDarcySkin(UnitDimension):
    """Non darcy skin units."""
    english = 'day/mscf'
    metric = 'day/stm3'
    metkgcm2 = 'day/stm3'
    metbar = 'day/stm3'
    lab = 'hour/stcc'
    metric_atm = 'day/stm3'


class ProductivityIndex(UnitDimension):
    """Productivity index units."""
    english = 'stb/day/psi'
    metric = 'stm3/day/kPa'
    metkgcm2 = 'stm3/day/kg/cm2'
    metbar = 'stm3/day/bars'
    lab = 'stcc/hour/psi'
    metric_atm = 'stm3/day/atm'


class ReservoirProductivityIndex(UnitDimension):
    """Productivity index reservoir units."""
    english = 'rb/day/psi'
    metric = 'm3/day/kPa'
    metkgcm2 = 'm3/day/kg/cm2'
    metbar = 'm3/day/bars'
    lab = 'cc/hour/psi'
    metric_atm = 'm3/day/atm'


class ReservoirVolumeOverPressure(UnitDimension):
    """Reservoir volume over pressure, for instance as needed for Carter-Tracy constant."""
    english = 'rb/psia'
    metric = 'm3/kPa'
    metkgcm2 = 'm3/kg/cm2'
    metbar = 'm3/bars'
    lab = 'cc/psia'
    metric_atm = 'stm3/day/atm'


class ValveCoefficient(UnitDimension):
    """Valve coefficient units."""
    english = 'psi(lb/ft3)/(lb/s)2'
    metric = 'kPa(kg/m3)/(kg/s)2'
    metkgcm2 = '(kg/cm2) (kg/m3)/(kg/s)2'
    metbar = 'bar(kg/m3)/(kg/s)2'
    lab = 'psi(gm/cc)/(gm/s)2'
    metric_atm = 'atm(kg/m3)/(kg/s)2'


class InterfacialTension(UnitDimension):
    """Interfacial tension units."""
    english = 'dynes/cm'
    metric = 'dynes/cm'
    metkgcm2 = 'dynes/cm'
    metbar = 'dynes/cm'
    lab = 'dynes/cm'
    metric_atm = 'dynes/cm'



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Sequence, Optional, TYPE_CHECKING
from uuid import UUID

import pandas as pd

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.File import File
if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import Network


class NetworkOperationsMixIn(ABC):
    def __init__(self, parent_network: Network) -> None:
        self.__parent_network = parent_network

    @abstractmethod
    def get_all(self) -> Sequence[Any]:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_by_name(self, name: str) -> Optional[Any]:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_df(self) -> pd.DataFrame:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def get_overview(self) -> str:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def load(self, file: File, start_date: str, default_units: UnitSystem) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def _add_to_memory(self, additional_objs: Optional[list[Any]]):
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def remove(self, obj_to_remove: UUID | dict[str, None | str | float | int]) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def add(self, obj_to_add: dict[str, None | str | float | int]) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @abstractmethod
    def modify(self, obj_to_modify: dict[str, None | str | float | int],
               new_properties: dict[str, None | str | float | int]) -> None:
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def table_header(self) -> str:
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def table_footer(self) -> str:
        raise NotImplementedError("Implement this in the derived class")

    @property
    @abstractmethod
    def _network_element_name(self) -> str:
        raise NotImplementedError("Implement this in the derived class")



from abc import ABC
from dataclasses import dataclass, field
from typing import Literal

from ResSimpy.OperationsMixin import NetworkOperationsMixIn
from ResSimpy.Wellhead import Wellhead


@dataclass(kw_only=True)
class Wellheads(NetworkOperationsMixIn, ABC):
    __wellheads: list[Wellhead] = field(default_factory=list)

    @property
    def _network_element_name(self) -> Literal['wellheads']:
        return 'wellheads'



from __future__ import annotations

import warnings
from typing import Optional, Sequence
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusWell import NexusWell
from ResSimpy.Nexus.DataModels.NexusCompletion import NexusCompletion
from ResSimpy.Nexus.DataModels.NexusRelPermEndPoint import NexusRelPermEndPoint
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.NexusKeywords.wells_keywords import WELLS_KEYWORDS


def load_wells(nexus_file: NexusFile, start_date: str, default_units: UnitSystem,
               model_date_format: DateFormat) -> tuple[Sequence[NexusWell], DateFormat]:
    """Loads a list of Nexus Well instances and populates it with the wells completions over time from a wells file.

    Args:
        nexus_file (NexusFile): NexusFile containing the wellspec files.
        start_date (str): starting date of the wellspec file as a string.
        default_units (UnitSystem): default units to use if no units are found.
        model_date_format (DateFormat): Date format specified in the FCS file.

    Raises:
        ValueError: If no value is found after a TIME card.
        ValueError: If no well name is found after a WELLSPEC keyword.
        ValueError: If no valid wells are found in the wellspec file.

    Returns:
        A tuple containing:
        Sequence[NexusWell]: list of Nexus well classes contained within a wellspec file.
        DateFormat: The date format found in the wellspec file if present, otherwise just the model date format.
    """
    date_format = model_date_format
    file_as_list = nexus_file.get_flat_list_str_file
    well_name: Optional[str] = None
    wellspec_file_units: Optional[UnitSystem] = None

    iw: Optional[str] = None
    jw: Optional[str] = None
    kw: Optional[str] = None
    md: Optional[str] = None
    skin: Optional[str] = None
    depth: Optional[str] = None
    x_value: Optional[str] = None
    y_value: Optional[str] = None
    angla: Optional[str] = None
    anglv: Optional[str] = None
    grid: Optional[str] = None
    wi: Optional[str] = None
    dtop: Optional[str] = None
    dbot: Optional[str] = None
    partial_perf: Optional[str] = None
    well_radius: Optional[str] = None

    perm_thickness_ovr: Optional[str] = None
    dfactor: Optional[str] = None
    rel_perm_method: Optional[str] = None
    status: Optional[str] = None

    cell_number: Optional[str] = None
    peaceman_well_block_radius: Optional[str] = None
    portype: Optional[str] = None
    fracture_mult: Optional[str] = None
    sector: Optional[str] = None
    well_group: Optional[str] = None
    zone: Optional[str] = None
    angle_open_flow: Optional[str] = None
    temperature: Optional[str] = None
    flowsector: Optional[str] = None
    parent_node: Optional[str] = None
    mdcon: Optional[str] = None
    pressure_avg_pattern: Optional[str] = None
    length: Optional[str] = None
    permeability: Optional[str] = None
    non_darcy_model: Optional[str] = None
    comp_dz: Optional[str] = None
    layer_assignment: Optional[str] = None
    polymer_bore_radius: Optional[str] = None
    polymer_well_radius: Optional[str] = None
    kh_mult: Optional[float] = None

    # End point values:
    swl: Optional[str] = None
    swr: Optional[str] = None
    swu: Optional[str] = None
    sgl: Optional[str] = None
    sgr: Optional[str] = None
    sgu: Optional[str] = None
    swro: Optional[str] = None
    sgro: Optional[str] = None
    sgrw: Optional[str] = None
    krw_swro: Optional[str] = None
    krw_swu: Optional[str] = None
    krg_sgro: Optional[str] = None
    krg_sgu: Optional[str] = None
    kro_swl: Optional[str] = None
    kro_swr: Optional[str] = None
    kro_sgl: Optional[str] = None
    kro_sgr: Optional[str] = None
    krw_sgl: Optional[str] = None
    krw_sgr: Optional[str] = None
    krg_sgrw: Optional[str] = None
    sgtr: Optional[str] = None
    sotr: Optional[str] = None
    header_values: dict[str, None | int | float | str] = {
        'IW': iw, 'JW': jw, 'L': kw, 'MD': md, 'SKIN': skin, 'DEPTH': depth, 'X': x_value, 'Y': y_value,
        'ANGLA': angla, 'ANGLV': anglv, 'GRID': grid, 'WI': wi, 'DTOP': dtop, 'DBOT': dbot, 'RADW': well_radius,
        'PPERF': partial_perf, 'CELL': cell_number, 'KH': perm_thickness_ovr, 'D': dfactor, 'IRELPM': rel_perm_method,
        'STAT': status, 'RADB': peaceman_well_block_radius, 'PORTYPE': portype, 'FM': fracture_mult, 'SECT': sector,
        'GROUP': well_group, 'ZONE': zone, 'ANGLE': angle_open_flow, 'TEMP': temperature, 'FLOWSECT': flowsector,
        'PARENT': parent_node, 'MDCON': mdcon, 'IPTN': pressure_avg_pattern, 'LENGTH': length, 'K': permeability,
        'ND': non_darcy_model, 'DZ': comp_dz, 'LAYER': layer_assignment, 'RADBP': polymer_bore_radius,
        'RADWP': polymer_well_radius, 'KHMULT': kh_mult
    }
    end_point_scaling_header_values: dict[str, None | int | float | str] = {
        'SWL': swl, 'SWR': swr, 'SWU': swu, 'SGL': sgl, 'SGR': sgr, 'SGU': sgu,
        'SWRO': swro, 'SGRO': sgro, 'SGRW': sgrw, 'KRW_SWRO': krw_swro, 'KRW_SWU': krw_swu, 'KRG_SGRO': krg_sgro,
        'KRG_SGU': krg_sgu, 'KRO_SWL': kro_swl, 'KRO_SWR': kro_swr, 'KRO_SGL': kro_sgl, 'KRO_SGR': kro_sgr,
        'KRW_SGL': krw_sgl, 'KRW_SGR': krw_sgr, 'KRG_SGRW': krg_sgrw, 'SGTR': sgtr, 'SOTR': sotr,
    }
    # add end point scaling header to the header values we search for:
    header_values.update(end_point_scaling_header_values)

    header_index: int = -1
    wellspec_found: bool = False
    current_date: Optional[str] = None
    wells: list[NexusWell] = []
    well_name_list: list[str] = []

    for index, line in enumerate(file_as_list):
        uppercase_line = line.upper()

        # If we haven't got the units yet, check to see if this line contains a declaration for them.
        if wellspec_file_units is None:
            for unit in UnitSystem:
                if unit.value in uppercase_line and (line.find('!') > line.find(unit.value) or line.find('!') == -1):
                    wellspec_file_units = unit

        if nfo.check_token('DATEFORMAT', line):
            new_date_format_str = nfo.get_token_value(token='DATEFORMAT', token_line=line, file_list=file_as_list)

            if new_date_format_str is None:
                raise ValueError(f"Cannot find the date format associated with the DATEFORMAT card in {line=} at line"
                                 f" number {index}")

            model_date_format_str = model_date_format.name.replace('_', '/')
            if new_date_format_str != model_date_format_str:
                warnings.warn(f"Wells date format of {new_date_format_str} inconsistent with base model format of "
                              f"{model_date_format_str}")

            converted_format_str = new_date_format_str.replace('/', '_')

            if not hasattr(DateFormat, converted_format_str):
                raise ValueError(f"Invalid Date Format found: '{new_date_format_str}' at line {index}")

            date_format = DateFormat[converted_format_str]

        if nfo.check_token('TIME', line):
            current_date = nfo.get_token_value(token='TIME', token_line=line, file_list=file_as_list)
            if current_date is None:
                raise ValueError(f"Cannot find the date associated with the TIME card in {line=} at line number \
                                 {index}")

        if nfo.check_token('WELLSPEC', uppercase_line):
            initial_well_name = nfo.get_expected_token_value(token='WELLSPEC', token_line=line, file_list=file_as_list,
                                                             custom_message="Cannot find well name following WELLSPEC "
                                                                            "keyword")
            well_name = initial_well_name.strip('\"')
            wellspec_found = True
            continue

        # Load in the column headings, which appear after the well name
        header_index, headers = __load_wellspec_table_headings(header_index, header_values, index, line, well_name)

        if header_index == -1:
            continue

        if well_name is None:
            raise ValueError(f"No wells found in file: {nexus_file.location}")

        if wellspec_found:
            if current_date is None:
                current_date = start_date
            # reset the storage dictionary to prevent completion properties being carried forward from earlier timestep
            header_values = {k: None for k in header_values}
            if wellspec_file_units is None:
                wellspec_file_units = default_units

            # Load in each line of the table
            completions = __load_wellspec_table_completions(
                nexus_file, header_index, header_values, headers, current_date, end_point_scaling_header_values,
                date_format, unit_system=wellspec_file_units)

            if well_name in well_name_list:
                wells[well_name_list.index(well_name)].completions.extend(completions)
            else:
                new_well = NexusWell(completions=completions, well_name=well_name, unit_system=wellspec_file_units)
                well_name_list.append(well_name)
                wells.append(new_well)
            wellspec_found = False
    return wells, date_format


def __load_wellspec_table_completions(nexus_file: NexusFile, header_index: int,
                                      header_values: dict[str, None | int | float | str],
                                      headers: list[str], start_date: str,
                                      end_point_scaling_header_values: dict[str, None | int | float | str],
                                      date_format: DateFormat,
                                      unit_system: UnitSystem,
                                      ) -> list[NexusCompletion]:
    """Loads a completion table in for a single WELLSPEC keyword. \
        Loads in the next available completions following a WELLSPEC keyword and a header line.

    Args:
        header_index (int): index number of the header in the file as list parameter
        header_values (dict[str, Union[Optional[int], Optional[float], Optional[str]]]): dictionary of column \
            headings to populate from the table
        headers (list[str]): list of strings containing the headers from the wellspec table
        start_date (str): date to populate the completion class with.

    Returns:
        list[NexusCompletion]: list of nexus completions for a given table.
    """

    def convert_header_value_float(key: str) -> Optional[float]:
        value = header_values[key]
        if value == 'NA':
            value = None
        if value is None:
            return None
        try:
            value = float(value)
        except ValueError:
            warnings.warn(f"Cannot convert {value=} to float")
            value = None
        return value

    def convert_header_value_int(key: str) -> Optional[int]:
        value = header_values[key]
        if value == 'NA':
            value = None
        return None if value is None else int(value)

    completions: list[NexusCompletion] = []
    file_as_list: list[str] = nexus_file.get_flat_list_str_file

    for index, line in enumerate(file_as_list[header_index + 1:]):
        # check for end of table lines:
        # TODO update with a more robust table end checker function
        end_of_table = nfo.nexus_token_found(line, WELLS_KEYWORDS)
        if end_of_table:
            return completions
        valid_line, header_values = nfo.table_line_reader(header_values, headers, line)
        # if a valid line is found load a completion otherwise continue
        if not valid_line:
            continue

        # create a rel perm end point scaling object if it exists for a given completion
        rel_perm_dict = {key.lower(): convert_header_value_float(key) for key
                         in header_values if key in end_point_scaling_header_values}
        if any(rel_perm_dict.values()):
            new_rel_perm_end_point = NexusRelPermEndPoint(**rel_perm_dict)
        else:
            new_rel_perm_end_point = None

        new_completion = NexusCompletion(
            date=start_date,
            i=convert_header_value_int('IW'),
            j=convert_header_value_int('JW'),
            k=convert_header_value_int('L'),
            skin=convert_header_value_float('SKIN'),
            depth=convert_header_value_float('DEPTH'),
            well_radius=convert_header_value_float('RADW'),
            x=convert_header_value_float('X'),
            y=convert_header_value_float('Y'),
            angle_a=convert_header_value_float('ANGLA'),
            angle_v=convert_header_value_float('ANGLV'),
            grid=(None if header_values['GRID'] is None else str(header_values['GRID'])),
            measured_depth=convert_header_value_float('MD'),
            well_indices=convert_header_value_float('WI'),
            depth_to_top=convert_header_value_float('DTOP'),
            depth_to_bottom=convert_header_value_float('DBOT'),
            depth_to_top_str=(None if header_values['DTOP'] is None else str(header_values['DTOP'])),
            depth_to_bottom_str=(None if header_values['DBOT'] is None else str(header_values['DBOT'])),
            rel_perm_method=convert_header_value_int('IRELPM'),
            dfactor=convert_header_value_float('D'),
            status=(None if header_values['STAT'] is None else str(header_values['STAT'])),
            partial_perf=convert_header_value_float('PPERF'),
            cell_number=convert_header_value_int('CELL'),
            perm_thickness_ovr=convert_header_value_float('KH'),
            peaceman_well_block_radius=convert_header_value_float('RADB'),
            fracture_mult=convert_header_value_float('FM'),
            sector=convert_header_value_int('SECT'),
            well_group=(None if header_values['GROUP'] is None else str(header_values['GROUP'])),
            zone=convert_header_value_int('ZONE'),
            angle_open_flow=convert_header_value_float('ANGLE'),
            temperature=convert_header_value_float('TEMP'),
            flowsector=convert_header_value_int('FLOWSECT'),
            parent_node=(None if header_values['PARENT'] is None else str(header_values['PARENT'])),
            mdcon=convert_header_value_float('MDCON'),
            pressure_avg_pattern=convert_header_value_int('IPTN'),
            length=convert_header_value_float('LENGTH'),
            permeability=convert_header_value_float('K'),
            non_darcy_model=(None if header_values['ND'] is None else str(header_values['ND'])),
            comp_dz=convert_header_value_float('DZ'),
            layer_assignment=convert_header_value_int('LAYER'),
            polymer_bore_radius=convert_header_value_float('RADBP'),
            polymer_well_radius=convert_header_value_float('RADWP'),
            portype=(None if header_values['PORTYPE'] is None else str(header_values['PORTYPE'])),
            rel_perm_end_point=new_rel_perm_end_point,
            kh_mult=convert_header_value_float('KHMULT'),
            date_format=date_format,
            unit_system=unit_system,
        )

        nexus_file.add_object_locations(new_completion.id, [index + header_index + 1])

        completions.append(new_completion)

    return completions


def __load_wellspec_table_headings(header_index: int, header_values: dict[str, None | int | float | str],
                                   index: int, line: str, well_name: Optional[str],
                                   headers: Optional[list[str]] = None) -> tuple[int, list[str]]:
    """Loads in the wellspec headers from a line in a file.

    Args:
        header_index (int): index of the header
        header_values (dict[str, Union[Optional[int], Optional[float], Optional[str]]]): dictionary of column \
            headings to populate from the table
        index (int): starting index to search from
        line (str): line to extract header values from
        well_name (Optional[str]): well name from the previous WELLSPEC keyword
        headers (Optional[list[str]], optional): list of headers to append the next set of found headers to. \
            Defaults to None and will create a new list to return if None.

    Returns:
        tuple[int, list[str]]: index in the file as list for the header, list of headers found in the file
    """
    headers = [] if headers is None else headers
    next_column_heading: Optional[str]
    if well_name is None:
        return header_index, headers

    for key in header_values.keys():
        if nfo.check_token(key, line):
            header_line = line.upper()
            header_index = index
            # Map the headers (first time get the expected value as check token guarantees at least 1 value)
            next_column_heading = nfo.get_expected_next_value(start_line_index=0, file_as_list=[line]).upper()
            trimmed_line = header_line

            while next_column_heading is not None:
                headers.append(next_column_heading)
                trimmed_line = trimmed_line.replace(next_column_heading, "", 1)
                next_column_heading = nfo.get_next_value(0, [trimmed_line], trimmed_line)

            if len(headers) > 0:
                break
    return header_index, headers



from __future__ import annotations
from abc import ABC
from dataclasses import dataclass
from typing import Optional

from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.DataObjectMixin import DataObjectMixin
from ResSimpy.Units.AttributeMappings.NetworkUnitMapping import NetworkUnits


@dataclass(repr=False)
class WellConnection(DataObjectMixin, ABC):
    name: Optional[str] = None
    date: Optional[str] = None
    unit_system: Optional[UnitSystem] = None

    bhdepth: Optional[float] = None
    datum_depth: Optional[float] = None
    x_pos: Optional[float] = None
    y_pos: Optional[float] = None
    length: Optional[float] = None
    temperature: Optional[float] = None
    diameter: Optional[float] = None
    roughness: Optional[float] = None
    inner_diameter: Optional[float] = None
    productivity_index: Optional[float] = None

    @property
    def units(self) -> NetworkUnits:
        """Returns the attribute to unit map for the WellConnection."""
        return NetworkUnits(self.unit_system)



# TODO: Not a complete list. Check manual for missing keywords.
# TODO: Delete the keywords that are not tokens (i.e. that are only values in a table)

WATER_KEYWORDS = ['BW', 'CW', 'DENW', 'PPM', 'PREF', 'SALINITY', 'SUNITS', 'TEMP', 'VISW']



"""Used to represent a connection between two nodes in a Nexus network."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.NodeConnection import NodeConnection


@dataclass(repr=False)
class NexusNodeConnection(NodeConnection):
    """Used to represent a connection between two nodes in a Nexus network.

    Attributes:
    hyd_method: hydraulic lift correlation method used (METHOD)
    pvt_method: pvt method number (IPVT)
    water_method: water method number (IWAT)
    bat_method: bat method number (IBAT)
    elevation_profile: elevation profile as a string (ELEVPR)
    measured_depth_in: measured depth of the input node (MDIN)
    measured_depth_out: measured depth of the output node (MDOUT)
    diameter: diameter of the connecting pipe (DIAMETER)
    inner_diameter: outer diameter of the pipe (INNERDIAM)
    roughness: roughness of the pipe (ROUGHNESS)
    heat_transfer_coeff: the heat transfer coefficient of the pipe (HTC)
    temperature_profile: temperature profile used for the connection. (TEMPPR)
    length: length of the connection (LENGTH)
    delta_depth: change in depth between node_in and node_out (DDEPTH)
    connection_number: used in place of name as a numbered connection (NUMBER)
    seawater_profile: seawater profile used. (SEAWPR)
    rate_mult: multiplier to the rate (RATEMULT)
    polymer: whether polymer is a stream here (POLYMER)
    dp_add: Additional delta pressure (DPADD).
    """

    hyd_method: Optional[str | int] = None
    pvt_method: Optional[int] = None
    bat_method: Optional[int] = None
    water_method: Optional[int] = None
    elevation_profile: Optional[str] = None
    measured_depth_in: Optional[float] = None
    measured_depth_out: Optional[float] = None
    diameter: Optional[float] = None
    inner_diameter: Optional[float] = None
    roughness: Optional[float] = None
    heat_transfer_coeff: Optional[float] = None
    temperature_profile: Optional[float] = None
    length: Optional[float] = None
    delta_depth: Optional[float] = None
    connection_number: Optional[int] = None
    seawater_profile: Optional[str] = None
    rate_mult: Optional[float] = None
    polymer: Optional[str] = None
    unit_system: Optional[UnitSystem] = None
    dp_add: Optional[float] = None
    dt_add: Optional[float] = None
    temperature_in: Optional[float] = None
    temperature_out: Optional[float] = None
    tracer: Optional[str] = None
    heat_conductivity: Optional[float] = None
    insulation_thickness: Optional[float] = None
    insulation_conductivity: Optional[float] = None
    gravity_pressure_gradient_mult: Optional[float] = None
    friction_pressure_gradient_mult: Optional[float] = None
    acceleration_pressure_gradient_mult: Optional[float] = None

    def __init__(self, properties_dict: dict[str, None | int | str | float]) -> None:
        # call the init of the DataObjectMixin
        super(NodeConnection, self).__init__({})
        for key, prop in properties_dict.items():
            self.__setattr__(key, prop)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        nexus_mapping = {
            'NAME': ('name', str),
            'NODEIN': ('node_in', str),
            'NODEOUT': ('node_out', str),
            'TYPE': ('con_type', str),
            'METHOD': ('hyd_method', str),
            'IPVT': ('pvt_method', int),
            'IWAT': ('water_method', int),
            'IBAT': ('bat_method', int),
            'ELEVPR': ('elevation_profile', str),
            'MDIN': ('measured_depth_in', float),
            'MDOUT': ('measured_depth_out', float),
            'DIAM': ('diameter', float),
            'INNERDIAM': ('inner_diameter', float),
            'ROUGHNESS': ('roughness', float),
            'HTC': ('heat_transfer_coeff', float),
            'TEMPPR': ('temperature_profile', str),
            'LENGTH': ('length', float),
            'DDEPTH': ('delta_depth', float),
            'NUMBER': ('connection_number', int),
            'SEAWPR': ('seawater_profile', str),
            'RATEMULT': ('rate_mult', float),
            'POLYMER': ('polymer', str),
            'DPADD': ('dp_add', float),
            'DTADD': ('dt_add', float),
            'TEMPIN': ('temperature_in', float),
            'TEMPOUT': ('temperature_out', float),
            'TRACERS': ('tracer', str),
            'HCOND': ('heat_conductivity', float),
            'INSTHN': ('insulation_thickness', float),
            'INSK': ('insulation_conductivity', float),
            'GRPGCR': ('gravity_pressure_gradient_mult', float),
            'FRPGCR': ('friction_pressure_gradient_mult', float),
            'ACPGCR': ('acceleration_pressure_gradient_mult', float),
        }
        return nexus_mapping



from abc import abstractmethod
from ResSimpy.Utils.invert_nexus_map import nexus_keyword_to_attribute_name
from typing import Protocol


class SupportsKeywordMapping(Protocol):
    @staticmethod
    @abstractmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        pass

    @abstractmethod
    def to_dict(self) -> dict[str, None | str | int | float]:
        pass


def to_table_line(obj: SupportsKeywordMapping, headers: list[str]) -> str:
    """Takes a generic Nexus object and returns the attribute values as a string in the order of headers provided.
    Requires an implemented to_dict method and get_keyword_mapping() method.

    Args:
        obj (Any): a NexusObject with a to_dict and get_keyword_mapping method.
        headers (list[str]): list of header values in Nexus keyword format

    Returns:
        string of the values in the order of the headers provided.

    """
    nexus_mapping = obj.get_keyword_mapping()
    properties = obj.to_dict()

    constructed_values = []
    for header in headers:
        attribute_name = nexus_keyword_to_attribute_name(nexus_mapping, header)
        attribute_value = properties[attribute_name]
        if attribute_value is None:
            attribute_value = 'NA'
        constructed_values.append(attribute_value)
    return ' '.join([str(x) for x in constructed_values]) + '\n'



"""Holds the NexusNodes class which is used to store and manipulate the nodes in a NexusNetwork.
It is stored as an instance in the NexusNetwork class as "nodes".
"""

from __future__ import annotations
from dataclasses import dataclass, field
from uuid import UUID
from typing import Sequence, Optional, TYPE_CHECKING

import pandas as pd

from ResSimpy.File import File
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.Nexus.nexus_collect_tables import collect_all_tables_to_objects
from ResSimpy.Nexus.DataModels.Network.NexusNode import NexusNode
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.nexus_modify_object_in_file import ModifyObjectOperations
from ResSimpy.Nexus.nexus_remove_object_from_file import RemoveObjectOperations
from ResSimpy.Nodes import Nodes
from ResSimpy.Utils.obj_to_dataframe import obj_to_dataframe

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork


@dataclass(kw_only=True)
class NexusNodes(Nodes):
    """Class to store and manipulate the nodes in a NexusNetwork.
    It is stored as an instance in the NexusNetwork class as "nodes".
    """
    __nodes: list[NexusNode] = field(default_factory=list)

    def __init__(self, parent_network: NexusNetwork) -> None:
        self.__parent_network: NexusNetwork = parent_network
        self.__nodes: list[NexusNode] = []
        self.__add_object_operations = AddObjectOperations(NexusNode, self.table_header, self.table_footer,
                                                           self.__parent_network.model)
        self.__remove_object_operations = RemoveObjectOperations(self.__parent_network, self.table_header,
                                                                 self.table_footer)
        self.__modify_object_operations = ModifyObjectOperations(self)

    @property
    def table_header(self) -> str:
        """Start of the Node definition table."""
        return 'NODES'

    @property
    def table_footer(self) -> str:
        """End of the Node definition table."""
        return 'ENDNODES'

    def get_all(self) -> Sequence[NexusNode]:
        """Returns a list of nodes loaded from the simulator."""
        self.__parent_network.get_load_status()
        return self.__nodes

    def get_by_name(self, node_name: str) -> Optional[NexusNode]:
        """Returns a single node with the provided name loaded from the simulator.

        Args:
        ----
            node_name (str): name of the requested node

        Returns:
        -------
            NexusNode: which has the same name as the requested node_name

        """
        nodes_to_return = filter(lambda x: False if x.name is None else x.name.upper() == node_name.upper(),
                                 self.__nodes)
        return next(nodes_to_return, None)

    def get_df(self) -> pd.DataFrame:
        """Creates a dataframe representing all processed node data in a surface file
        Returns:
            DataFrame: of the properties of the nodes through time with each row representing a node.
        """
        df_store = obj_to_dataframe(self.__nodes)
        return df_store

    def get_overview(self) -> str:
        raise NotImplementedError('To be implemented')

    def load(self, surface_file: File, start_date: str, default_units: UnitSystem) -> None:
        """Calls load nodes and appends the list of discovered nodes into the NexusNodes object.

        Args:
            surface_file (File): NexusFile representation of the surface file.
            start_date (str): Starting date of the run
            default_units (UnitSystem): Units used in case not specified by surface file.

        Raises:
            TypeError: if the unit system found in the property check is not a valid enum UnitSystem.

        """
        new_nodes = collect_all_tables_to_objects(surface_file, {'NODES': NexusNode},
                                                  start_date=start_date,
                                                  default_units=default_units)
        cons_list = new_nodes.get('NODES')
        if isinstance(cons_list, dict):
            raise ValueError('Incompatible data format for additional wells. Expected type "list" instead got "dict"')
        self._add_to_memory(cons_list)

    def _add_to_memory(self, additional_list: Optional[list[NexusNode]]) -> None:
        """Extends the nodes object by a list of nodes provided to it.

        Args:
        ----
            additional_list (Sequence[NexusNode]): list of nexus nodes to add to the nodes list.

        Returns:
        -------
            None
        """
        if additional_list is None:
            return
        self.__nodes.extend(additional_list)

    def remove(self, node_to_remove: dict[str, None | str | float | int] | UUID) -> None:
        """Remove a node from the network based on the properties matching a dictionary or id.

        Args:
            node_to_remove (UUID | dict[str, None | str | float | int]): UUID of the node to remove or a dictionary \
            with sufficient matching parameters to uniquely identify a node

        """
        self.__remove_object_operations.remove_object_from_network_main(
            node_to_remove, self._network_element_name, self.__nodes)

    def add(self, node_to_add: dict[str, None | str | float | int]) -> None:
        """Adds a node to a network, taking a dictionary with properties for the new node.

        Args:
            node_to_add (dict[str, None | str | float | int]): dictionary taking all the properties for the new node.
            Requires date and a node name.
        """
        new_object = self.__add_object_operations.add_network_obj(node_to_add, NexusNode, self.__parent_network)
        self._add_to_memory([new_object])

    def modify(self, node_to_modify: dict[str, None | str | float | int],
               new_properties: dict[str, None | str | float | int]) -> None:
        """Modifies an existing node based on a matching dictionary of properties (partial matches allowed if precisely
         1 matching node is found). Updates the properties with properties in the new_properties dictionary.

        Args:
            node_to_modify (dict[str, None | str | float | int]): dictionary containing attributes to match in the
            existing node set.
            new_properties (dict[str, None | str | float | int]): properties to switch to in the new node
        """
        self.__parent_network.get_load_status()

        self.__modify_object_operations.modify_network_object(node_to_modify, new_properties,
                                                              self.__parent_network)



from __future__ import annotations
from dataclasses import dataclass
import os
from typing import Optional, MutableMapping
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusHydraulicsMethod import NexusHydraulicsMethod
from ResSimpy.Hydraulics import Hydraulics


@dataclass(kw_only=True)
class NexusHydraulicsMethods(Hydraulics):
    """Class for collection of Nexus hydraulics methods.

    Attributes:
        inputs (dict[int, NexusHydraulicsMethod]): Collection of Nexus hydraulics methods, as a dictionary
        files (dict[int, NexusFile]): Dictionary collection of hydraulics files, in Nexus fcs file.
    """

    __inputs: MutableMapping[int, NexusHydraulicsMethod]
    __files: dict[int, NexusFile]
    __properties_loaded: bool = False  # Used in lazy loading
    __model_unit_system: UnitSystem

    def __init__(self, model_unit_system: UnitSystem,
                 inputs: Optional[MutableMapping[int, NexusHydraulicsMethod]] = None,
                 files: Optional[dict[int, NexusFile]] = None) -> None:
        if inputs:
            self.__inputs = inputs
        else:
            self.__inputs: MutableMapping[int, NexusHydraulicsMethod] = {}
        if files:
            self.__files = files
        else:
            self.__files = {}
        self.__model_unit_system = model_unit_system
        super().__init__()

    def __repr__(self) -> str:
        """Pretty printing hydraulics methods."""
        if not self.__properties_loaded:
            self.load_hydraulics_methods()
        printable_str = ''
        for table_num in self.__inputs.keys():
            printable_str += '\n--------------------------------\n'
            printable_str += f'HYD method {table_num}\n'
            printable_str += '--------------------------------\n'
            printable_str += self.__inputs[table_num].__repr__()
            printable_str += '\n'

        return printable_str

    @property
    def inputs(self) -> MutableMapping[int, NexusHydraulicsMethod]:
        if not self.__properties_loaded:
            self.load_hydraulics_methods()
        return self.__inputs

    @property
    def files(self) -> dict[int, NexusFile]:
        return self.__files

    def load_hydraulics_methods(self):
        # Read in hydraulics properties from Nexus hydraulics method files
        if self.__files is not None and len(self.__files) > 0:  # Check if hydraulics files exist
            for table_num in self.__files.keys():  # For each hydraulics property method
                hydraulics_file = self.__files[table_num]
                if hydraulics_file.location is None:
                    raise ValueError(f'Unable to find hydraulics file: {hydraulics_file}')
                if os.path.isfile(hydraulics_file.location):
                    # Create NexusHydraulicsMethod object
                    self.__inputs[table_num] = NexusHydraulicsMethod(file=hydraulics_file, input_number=table_num,
                                                                     model_unit_system=self.__model_unit_system)
                    self.__inputs[table_num].read_properties()  # Populate object with hydraulics props
        self.__properties_loaded = True



from abc import ABC
from dataclasses import dataclass, field
from typing import Literal

from ResSimpy.NodeConnection import NodeConnection
from ResSimpy.OperationsMixin import NetworkOperationsMixIn


@dataclass(kw_only=True)
class NodeConnections(NetworkOperationsMixIn, ABC):
    __connections: list[NodeConnection] = field(default_factory=list)

    @property
    def _network_element_name(self) -> Literal['connections']:
        return 'connections'



"""Creates a repr for data objects that removes attributes that are None from the representation."""
from typing import Any


def generic_repr(input_class: Any) -> str:
    """Creates a prettier object representation while removing attributes that are None from that
    representation.

    Args:
    ----
        input_class (Any): a class with attributes to summarise

    Returns:
    -------
        (str): Pretty representation of the string.
    """
    filtered_attrs = {k: v for k, v in vars(input_class).items() if v is not None}
    attrs = ', '.join(f"{k}={v!r}" for k, v in filtered_attrs.items())
    return f"{input_class.__class__.__name__}({attrs})"


def generic_str(input_class: Any) -> str:
    """Creates a string representation of the object while removing None and ids from the string.

    Args:
        input_class (Any): a class with attributes to summarise

    Returns:
        (str): String representation of the object.
    """
    filtered_attrs = {k: v for k, v in vars(input_class).items() if v is not None}
    id_keys = [id_attr for id_attr in filtered_attrs if id_attr.endswith('__id')]
    for id_attr in id_keys:
        del filtered_attrs[id_attr]

    attrs = ', '.join(f"{k}={v!r}" for k, v in filtered_attrs.items())
    return f"{input_class.__class__.__name__}({attrs})"



"""Class to hold data input for a Nexus Separator method."""
from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Union
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Nexus.NexusKeywords.separator_keywords import SEPARATOR_KEYS_INT, SEPARATOR_KEYS_FLOAT
from ResSimpy.Nexus.NexusKeywords.separator_keywords import SEPARATOR_KEYWORDS
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import SeparatorUnits
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusSeparatorMethod(DynamicProperty):
    """Class to hold data input for a Nexus Separator method.

    Attributes:
        file (NexusFile): Nexus Separator file object
        input_number (int): Separator method number in Nexus fcs file
        separator_type (Optional[str]): Type of separator method, e.g., BLACKOIL, GASPLANT or EOS. Defaults to None
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific separator method. Defaults to empty dictionary.
    """

    file: NexusFile
    separator_type: Optional[str] = None
    properties: dict[str, Union[str, int, float, Enum, list[str],
                     pd.DataFrame, dict[str, Union[float, pd.DataFrame]]]] \
        = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 separator_type: Optional[str] = None,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        if separator_type is not None:
            self.separator_type = separator_type
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'TEMP': ('temperature', float),
            'PRES': ('pressure', float),
            'TEMP_STD': ('standard_temperature', float),
            'PRES_STD': ('standard_pressure', float),
        }
        return keywords

    @property
    def units(self) -> SeparatorUnits:
        """Returns the attribute to unit map for the separator method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return SeparatorUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with separator data in Nexus file format."""
        printable_str = ''
        sep_dict = self.properties
        for key, value in sep_dict.items():
            if key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif key == 'SEPARATOR_TABLE' and isinstance(value, pd.DataFrame):
                if self.separator_type == 'GASPLANT':
                    printable_str += 'RECFAC_TABLE\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n'
                if self.separator_type == 'GASPLANT':
                    printable_str += 'ENDRECFAC_TABLE\n'
                printable_str += '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key == 'KEYCPTLIST' and isinstance(value, list):
                printable_str += f"{key} {' '.join(value)}\n"
            elif value == '':
                printable_str += f'{key}\n'
            else:
                printable_str += f'{key} {value}\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus Separator file contents and populate NexusSeparatorMethod object."""
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers used to record properties as we iterate through separator file contents
        # List to record start and ending indices for tables
        sep_table_indices: list[int] = [0, len(file_as_list)]
        # Flag to tell when to start reading a table
        start_reading_table: bool = False

        line_indx = 0
        for line in file_as_list:

            # Determine separator type, i.e., EOS multistage, gas plant data or black oil
            if nfo.check_token('TEMP', line):  # EOS multistage separator table
                self.separator_type = 'EOS'
                sep_table_indices[0] = line_indx  # Specify EOS multistage separator table starting index
                start_reading_table = True
                line_indx += 1
                continue
            elif nfo.check_token('KEYCPTLIST', line):  # Gas plant data
                self.separator_type = 'GASPLANT'
                elems = line.split('!')[0].split()
                cpt_index = elems.index('KEYCPTLIST')
                if 'KEYCPTLIST' not in self.properties.keys():
                    self.properties['KEYCPTLIST'] = elems[cpt_index + 1:]
                line_indx += 1
                continue
            elif nfo.check_token('BOSEP', line):  # Black oil separator table
                self.separator_type = 'BLACKOIL'
                self.properties['BOSEP'] = ''
                line_indx += 1
                continue

            # Find SEPARATOR key-value pairs, such as WATERMETHOD 1 or PRES_STD 14.7
            if [i for i in line.split() if i in SEPARATOR_KEYS_FLOAT]:
                for key in SEPARATOR_KEYS_FLOAT:
                    if nfo.check_token(key, line):
                        self.properties[key] = float(nfo.get_expected_token_value(key, line, file_as_list))
            if [i for i in line.split() if i in SEPARATOR_KEYS_INT]:
                for key in SEPARATOR_KEYS_INT:
                    if nfo.check_token(key, line):
                        self.properties[key] = int(nfo.get_expected_token_value(key, line, file_as_list))

            # Find starting index for black oil separator table
            if self.separator_type == 'BLACKOIL':
                if nfo.check_token('KVOIL', line):
                    sep_table_indices[0] = line_indx
                    start_reading_table = True
                    line_indx += 1
                    continue

            # Find ending index for EOS multistage and black oil separator tables
            if start_reading_table:
                if self.separator_type in ['EOS', 'BLACKOIL']:
                    for keyword in SEPARATOR_KEYWORDS:
                        if nfo.check_token(keyword, line):
                            sep_table_indices[1] = line_indx
                            start_reading_table = False
                            break

            # Find starting and ending indices for gas plant separator table
            if self.separator_type == 'GASPLANT':
                if nfo.check_token('RECFAC_TABLE', line):
                    sep_table_indices[0] = line_indx + 1
                    start_reading_table = True
                if nfo.check_token('ENDRECFAC_TABLE', line):
                    sep_table_indices[1] = line_indx
                    start_reading_table = False

            line_indx += 1

        # Read in separator table
        if self.separator_type is not None:
            self.properties['SEPARATOR_TABLE'] = nfo.read_table_to_df(file_as_list[sep_table_indices[0]:
                                                                                   sep_table_indices[1]])



from __future__ import annotations
from dataclasses import dataclass, field
from abc import ABC
from uuid import UUID

import pandas as pd
from ResSimpy.Enums.HowEnum import OperationEnum

from ResSimpy.Well import Well
from typing import Sequence, Optional


@dataclass(kw_only=True)
class Wells(ABC):
    _wells: Sequence[Well] = field(default_factory=list)
    _wells_loaded: bool = False

    def __init__(self, assume_loaded: bool = False) -> None:
        self._wells_loaded = assume_loaded
        self._wells = []

    @property
    def wells(self) -> Sequence[Well]:
        if not self._wells_loaded:
            self._load()
        return self._wells

    @wells.setter
    def wells(self, value) -> None:
        self._wells = value

    def _load(self) -> None:
        raise NotImplementedError("Implement this in the derived class")

    def get_all(self) -> Sequence[Well]:
        raise NotImplementedError("Implement this in the derived class")

    def get(self, well_name: str) -> Optional[Well]:
        raise NotImplementedError("Implement this in the derived class")

    def get_df(self) -> pd.DataFrame:
        raise NotImplementedError("Implement this in the derived class")

    def modify(self, well_name: str, completion_properties_list: list[dict[str, None | float | int | str]],
               how: OperationEnum = OperationEnum.ADD) -> None:
        raise NotImplementedError("Implement this in the derived class")

    def add_completion(self, well_name: str, completion_properties: dict[str, None | float | int | str],
                       preserve_previous_completions: bool = True, comments: Optional[str] = None) -> None:
        raise NotImplementedError("Implement this in the derived class")

    def remove_completion(self, well_name: str,
                          completion_properties: Optional[dict[str, None | float | int | str]] = None,
                          completion_id: Optional[UUID] = None) -> None:
        raise NotImplementedError("Implement this in the derived class")

    def modify_completion(self, well_name: str, properties_to_modify: dict[str, None | float | int | str],
                          completion_to_change: Optional[dict[str, None | float | int | str]] = None,
                          completion_id: Optional[UUID] = None,
                          comments: Optional[str] = None) -> None:
        raise NotImplementedError("Implement this in the derived class")

    def get_wells_overview(self) -> str:
        if not self._wells_loaded:
            self._load()

        overview: str = ''
        for well in self._wells:
            overview += well.printable_well_info

        return overview

    def get_wells_dates(self) -> set[str]:
        """Returns a set of the unique dates in the wellspec file over all wells."""
        if not self._wells_loaded:
            self._load()

        set_dates: set[str] = set()
        for well in self._wells:
            set_dates.update(set(well.dates_of_completions))

        return set_dates

    @property
    def table_header(self) -> str:
        raise NotImplementedError("Implement this in the derived class")

    @property
    def table_footer(self) -> str:
        raise NotImplementedError("Implement this in the derived class")



from typing import Any

from ResSimpy.Enums.UnitsEnum import UnitSystem


def to_dict(nexus_object: Any, keys_in_nexus_style: bool = False, add_date: bool = True, add_units: bool = True,
            include_nones: bool = True) -> dict[str, None | str | int | float]:
    """Returns a dictionary of the attributes of a Nexus object. Requires a nexus mapping dictionary.
        Useful for creating dataframes of objects.

    Args:
    ----
        nexus_object (Any): Nexus object with a mapping dictionary defined
        keys_in_nexus_style (bool): if True returns the key values in Nexus keywords, otherwise returns the
            attribute name as stored by ressimpy
        add_date (bool): adds a date attribute if it exists
        add_units (bool): adds a units attribute if it exists.
        include_nones (bool): If False filters the nones out of the dictionary. Defaults to True

    Returns:
    -------
        a dictionary keyed by attributes and values as the value of the attribute
    """
    mapping_dict = nexus_object.get_keyword_mapping()
    if keys_in_nexus_style:
        result_dict = {x: nexus_object.__getattribute__(y[0]) for x, y in mapping_dict.items()}

    else:
        result_dict = {y[0]: nexus_object.__getattribute__(y[0]) for y in mapping_dict.values()}

    if add_date:
        try:
            result_dict['date'] = getattr(nexus_object, 'date')
        except AttributeError:
            raise AttributeError('Date was requested from the object but does not have a date associated with it.'
                                 f'Try setting add_date to False. Full contents of object: {nexus_object}')
    if add_units:
        try:
            unit_sys = getattr(nexus_object, 'unit_system')
        except AttributeError:
            raise AttributeError(
                'Unit system was requested from the object but does not have a unit system associated with it.'
                f'Try setting add_units to False. Full contents of the object: {nexus_object}')
        if isinstance(unit_sys, UnitSystem):
            result_dict['unit_system'] = unit_sys.value

    if not include_nones:
        result_dict = {k: v for k, v in result_dict.items() if v is not None}
    return result_dict



from dataclasses import dataclass


@dataclass
class FcsConfig:
    def __init__(self, destination, nexus_data_name="data") -> None:
        self.output_dir = destination
        self.use_reservoir_names = True
        self.abs_paths_to_keep: list[str] = []
        self.use_one_dir = True
        self.input_data_dir = ''
        self.path_components_to_keep = None
        self.locals_to_data = True
        self.output_data_dir = nexus_data_name
        self.dry_run = False
        self.compress = False
        self.output_log_file = False



from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
import re
from typing import Optional, Union
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.hyd_keywords import HYD_ARRAY_KEYWORDS, HYD_TABLE_HEADER_COLS, HYD_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.hyd_keywords import HYD_PRESSURE_KEYWORDS, HYD_SINGLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.hyd_keywords import HYD_KEYWORDS_VALUE_FLOAT, HYD_WATINJ_KEYWORDS_VALUE_FLOAT
from ResSimpy.Nexus.NexusKeywords.hyd_keywords import HYD_ALQ_KEYWORD, HYD_ALQ_OPTIONS
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import HydraulicsUnits

from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusHydraulicsMethod(DynamicProperty):
    """Class to hold Nexus Hydraulics properties.

    Attributes:
        file (NexusFile): Nexus hydraulics properties file object
        input_number (int): Hydraulics properties method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific hydraulics properties method. Defaults to empty dictionary.
    """

    # General parameters
    file: NexusFile
    properties: dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                     dict[str, Union[float, pd.DataFrame]]]] = field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem
    ratio_thousands: bool

    def __init__(self, file: NexusFile, input_number: int, model_unit_system: UnitSystem,
                 ratio_thousands: bool = True,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        self.ratio_thousands = ratio_thousands
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'QOIL': ('surface_oil_rate', float),
            'QLIQ': ('surface_liquid_rate', float),
            'QWAT': ('surface_water_rate', float),
            'QGAS': ('surface_gas_rate', float),
            'QWGAS': ('surface_wet_gas_rate', float),
            'MMW': ('mean_molecular_weight', float),
            'WCUT': ('watercut', float),
            'OCUT': ('oilcut', float),
            'PIN': ('inlet_pressure', float),
            'POUT': ('outlet_pressure', float),
            'BHP': ('bottomhole_pressure', float),
            'THP': ('tubinghead_pressure', float),
            'GLR': ('gas_liquid_ratio', float),
            'GOR': ('gas_oil_ratio', float),
            'OGR': ('oil_gas_ratio', float),
            'WGR': ('water_gas_ratio', float),
            'GWR': ('gas_water_ratio', float),
            'WWGR': ('water_wet_gas_ratio', float),
            'LENGTH': ('length', float),
            'DATUM': ('datum_depth', float),
            'DZ': ('depth_change', float),
            'DZW': ('hydraulic_table_vertical_distance', float),
            'GRAD': ('injected_fluid_pressure_gradient', float),
            'VISC': ('viscosity', float),
            'DIAM': ('diameter', float),
            'ROUGHNESS': ('roughness', float),
        }
        return keywords

    @property
    def units(self) -> HydraulicsUnits:
        """Returns the attribute to unit map for the Hydraulics method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return HydraulicsUnits(unit_system=self.unit_system, ratio_thousands=self.ratio_thousands)

    def to_string(self) -> str:
        """Create string with hydraulics data in Nexus file format."""
        printable_str = ''
        hyd_dict = self.properties
        for key, value in hyd_dict.items():
            if isinstance(value, pd.DataFrame):
                if key not in ['HYD_TABLE']:
                    printable_str += f'{key}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n'
                if key == 'LIMITS':
                    printable_str += 'ENDLIMITS\n'
                printable_str += '\n'
            elif key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key == 'ALQ':
                printable_str += f'{key}'
                if 'ALQ_PARAM' in hyd_dict.keys():
                    printable_str += f" {hyd_dict['ALQ_PARAM']}"
                printable_str += f' {value}\n'
            elif key == 'WATINJ' and isinstance(value, dict):
                printable_str += f'{key}\n'
                for watinj_key in value.keys():
                    printable_str += f'    {watinj_key} {value[watinj_key]}\n'
            elif key not in ['ALQ', 'ALQ_PARAM', 'WATINJ']:
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
        printable_str += '\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus hydraulics file contents and populate the NexusHydraulicsMethod object."""
        file_as_list0 = self.file.get_flat_list_str_file

        # Loop through file_as_list and fix any line continuations that use '>' at the end of the line
        file_as_list: list[str] = []
        if len(file_as_list0) > 0:
            file_as_list = [file_as_list0[0]]
            line_indx = 1
            while line_indx < len(file_as_list0):
                prev_line = file_as_list[-1]
                line = file_as_list0[line_indx]
                if prev_line.endswith(('>', '>\n')):
                    file_as_list[-1] = prev_line.split('>')[0] + ' ' + line.strip()
                else:
                    file_as_list.append(line)
                line_indx += 1

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers to use to record properties as we iterate through aquifer file contents
        # Dictionary to record start and ending indices for tables
        hyd_table_indices: dict[str, list[int]] = {}
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {'HYD_TABLE': False, 'LIMITS': False}
        start_reading_table = False
        # Containers for handling WATERINJ data
        watinj_dict: dict[str, float] = {}
        found_waterinj = False

        line_indx = 0
        for line in file_as_list:

            # Find arrays of parameters, e.g., QOIL 1.0 10. 100., or GOR 0.0 0.5 1.0
            potential_keyword = nfo.check_list_tokens(HYD_ARRAY_KEYWORDS + HYD_ALQ_KEYWORD, line)
            if potential_keyword is not None and not table_being_read['LIMITS']:
                line_elems = line.split('!')[0].split()
                next_val = nfo.get_expected_token_value(potential_keyword, line, file_as_list)
                if potential_keyword == 'ALQ' and next_val in HYD_ALQ_OPTIONS:
                    self.properties['ALQ_PARAM'] = next_val
                    keyword_index = line_elems.index(next_val)
                else:
                    keyword_index = line_elems.index(potential_keyword)
                self.properties[potential_keyword] = ' '.join(line_elems[keyword_index+1:])

            # Create WATINJ property if needed
            if nfo.check_token('WATINJ', line):
                found_waterinj = True

            # Handle DATGRAD property
            if nfo.check_token('DATGRAD', line):
                self.properties['DATGRAD'] = nfo.get_expected_token_value('DATGRAD', line, file_as_list)

            # Find HYD key-value pairs, such as LENGTH 3000, DATUM 7000 or DATGRAD GRAD
            potential_keyword = nfo.check_list_tokens(HYD_KEYWORDS_VALUE_FLOAT + HYD_WATINJ_KEYWORDS_VALUE_FLOAT, line)
            if potential_keyword is not None:
                if found_waterinj and potential_keyword in HYD_WATINJ_KEYWORDS_VALUE_FLOAT:
                    watinj_dict[potential_keyword] = float(
                        nfo.get_expected_token_value(potential_keyword, line, file_as_list))
                elif potential_keyword in HYD_KEYWORDS_VALUE_FLOAT:
                    self.properties[potential_keyword] = float(
                        nfo.get_expected_token_value(potential_keyword, line, file_as_list))

            # Find standalone hydraulics keywords
            potential_keyword = nfo.check_list_tokens(HYD_SINGLE_KEYWORDS, line)
            if potential_keyword is not None:
                self.properties[potential_keyword] = ''

            # Find starting and ending indices for hydraulic limits table
            if nfo.check_token('LIMITS', line):
                hyd_table_indices['LIMITS'] = [line_indx+1, len(file_as_list)]
                table_being_read['LIMITS'] = True
                start_reading_table = True
            if nfo.check_token('ENDLIMITS', line):
                hyd_table_indices['LIMITS'][1] = line_indx
                table_being_read['LIMITS'] = False
                start_reading_table = False

            # Find ending index of hydraulics table
            if start_reading_table and table_being_read['HYD_TABLE']:
                for potential_endkeyword in HYD_KEYWORDS:
                    if nfo.check_token(potential_endkeyword, line):
                        hyd_table_indices['HYD_TABLE'][1] = line_indx
                        start_reading_table = False
                        table_being_read['HYD_TABLE'] = False
                        break
            # Find starting index of hydraulics table
            for header_keyword in HYD_TABLE_HEADER_COLS:
                if nfo.check_token(header_keyword, line):
                    hyd_table_indices['HYD_TABLE'] = [line_indx, len(file_as_list)]
                    table_being_read['HYD_TABLE'] = True
                    start_reading_table = True

            line_indx += 1

        # Read in hydraulics tables
        for key in hyd_table_indices.keys():
            if key == 'HYD_TABLE':
                # Get hydraulic table data
                hyd_df: pd.DataFrame = nfo.read_table_to_df(file_as_list[
                    hyd_table_indices[key][0]+1:hyd_table_indices[key][1]], noheader=True)
                # Resolve hydraulic table header
                hyd_df_header_names = [header_name for header_name in file_as_list[
                    hyd_table_indices[key][0]].split('!')[0].split() if not re.search('BHP', header_name)]
                pressure_keyword_map = {'PIN': 'POUT', 'POUT': 'PIN', 'THP': 'BHP'}
                hyd_table_pressure_key = ''  # Determines what type of pressures are in table, i.e., PIN, POUT or BHP
                for pkey in HYD_PRESSURE_KEYWORDS:
                    if pkey in self.properties.keys():
                        if pkey == 'THP':
                            # For VIP compatibility, indicates units
                            # for GOR/GLR are SCF/STB rather than MSCF/STB
                            self.ratio_thousands = False
                        hyd_table_pressure_key = pressure_keyword_map[pkey]
                        break
                hyd_df_header_names = hyd_df_header_names + \
                    [hyd_table_pressure_key+str(i) for i in range(hyd_df.shape[1] - len(hyd_df_header_names))]
                hyd_df.columns = hyd_df_header_names
                # Set appropriate hydraulic table property
                self.properties[key] = hyd_df
            else:
                self.properties[key] = nfo.read_table_to_df(file_as_list[
                    hyd_table_indices[key][0]:hyd_table_indices[key][1]])

        # Correctly set WATINJ properties, if needed
        if found_waterinj:
            self.properties['WATINJ'] = watinj_dict



from abc import ABC
from dataclasses import dataclass
from typing import Sequence

from ResSimpy.Grid import Grid


@dataclass(kw_only=True)
class Grids(ABC):
    def get_grids(self) -> Sequence[Grid]:
        raise NotImplementedError("Implement this in the derived class")

    def load_grids(self) -> Sequence[Grid]:
        raise NotImplementedError("Implement this in the derived class")



"""Class for handling adding objects to a nexus file and memory."""
from __future__ import annotations
import copy
from typing import TYPE_CHECKING, Any, TypeVar, Optional
from uuid import UUID

from ResSimpy.File import File
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.DataObjectMixin import DataObjectMixin
from ResSimpy.Utils.invert_nexus_map import invert_nexus_map

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator
    from ResSimpy.Nexus.NexusNetwork import NexusNetwork

T = TypeVar('T', bound=DataObjectMixin)


class AddObjectOperations:
    """Class for handling adding objects to a nexus file and memory."""

    def __init__(self, obj_type: Optional[type[T]], table_header: str, table_footer: str,
                 model: NexusSimulator) -> None:
        self.__model = model
        self.table_header = table_header
        self.table_footer = table_footer
        self.obj_type = obj_type

    def find_which_file_from_id(self, obj_id: UUID, file_type_to_search: str) -> NexusFile:
        """Finds a file based on the presence of an object in the file.

        Args:
            obj_id (UUID): object id to match on
            file_type_to_search (str): file type from within the fcsfile. e.g. 'well_files'

        Returns:
            NexusFile that contains the object of interest in the object_locations attribute.
        """
        # find the correct wellspec file in the model by looking at the ids

        if self.__model.model_files is None:
            raise ValueError(f'File found in fcs file at: {self.__model.model_files}')
        files_to_search = getattr(self.__model.model_files, file_type_to_search, None)
        if files_to_search is None:
            raise FileNotFoundError(f'No files of type {files_to_search} found in model.')
        elif isinstance(files_to_search, NexusFile):
            # handling only 1 file returned of the file_type_to_search
            all_files = [files_to_search] if files_to_search.object_locations is not None and \
                                             obj_id in files_to_search.object_locations else []
        else:
            # handle case of multiple files of the same type.
            all_files = [x for x in files_to_search.values() if x.object_locations is not None and
                         obj_id in x.object_locations]
        if len(all_files) == 0:
            raise FileNotFoundError(f'No {file_type_to_search} found with and object with id: {obj_id}')
        elif len(all_files) > 1:
            raise ValueError(f'Too many files found containing that object id.'
                             f'Check if there are conflicts in where the objects are being stored.'
                             f'Files found matching: {[x.location for x in all_files]}')
        file_found = all_files[0]
        if file_found.file_content_as_list is None:
            raise FileNotFoundError(
                f'No well file content found for specified wellfile at location: {file_found.location}')
        return file_found

    @staticmethod
    def get_and_write_new_header(additional_headers: list[str],
                                 object_properties: dict[str, None | str | float | int],
                                 file_content: list[str], index: int,
                                 nexus_mapping: dict[str, tuple[str, type]], file: File) -> \
            tuple[int, list[str], list[str]]:
        """Gets the header and works out if any additional headers should be added based on the new obj attributes\
        being added.

        Args:
            additional_headers (list[str]): New headers required for the objects new columns.
            object_properties (dict[str, None | str | float | int]): dictionary containing new attributes of the object.
            file_content (list[str]): flattened file content as a list of strings.
            index (int): index to start reading the new headers from.
            nexus_mapping (dict[str, tuple[str, type]]): dictionary controlling the mapping from keyword to attribute.
            file (NexusFile): file to write the new header to.

        Returns:
            tuple[int, list[str], list[str]]: index of the new headers, a list of the headers,
            the original set of headers.
        """
        # TODO move out the additional headers mutability to a separate method that explicitly sets it
        keyword_map = {x: y[0] for x, y in nexus_mapping.items()}
        inverted_nexus_map = invert_nexus_map(nexus_mapping)
        table = file_content[index::]
        header_index, headers = nfo.get_table_header(file_as_list=table, header_values=keyword_map)
        header_index += index
        headers_original = copy.copy(headers)
        for key in object_properties:
            if key == 'date' or key == 'unit_system' or key == 'date_format':
                continue
            if inverted_nexus_map[key] not in headers:
                headers.append(inverted_nexus_map[key])
                additional_headers.append(inverted_nexus_map[key])
        additional_column_string = ' '.join(additional_headers)
        split_comments = str(file_content[header_index]).split('!', 1)
        if len(split_comments) == 1:
            new_header_line = split_comments[0].replace('\n', '', 1) + ' ' + additional_column_string + '\n'
        else:
            new_header_line = split_comments[0] + additional_column_string + ' !' + split_comments[1]
        if len(additional_headers) > 0:
            file_to_write_to, index_in_file = file.find_which_include_file(header_index)
            if file_to_write_to.file_content_as_list is None:
                raise ValueError(
                    f'No file content found in {file_to_write_to.location}. Cannot write to index {index_in_file}')
            file_to_write_to.file_content_as_list[index_in_file] = new_header_line
        return header_index, headers, headers_original

    @staticmethod
    def fill_in_nas(additional_headers: list[str], headers_original: list[str], index: int, line: str,
                    file: File, file_content: list[str]) -> int:
        """Check the validity of the line, if its valid add as many NAs as required for the new columns.

        Args:
            additional_headers (list[str]): New headers added to the table.
            headers_original (list[str]): headers from the original table in the file.
            index (int): index of the line to start filling NAs from.
            line (str): line to add additional NAs to.
            file (NexusFile): file to add NAs to
            file_content (list[str]): flattened content of the file being read

        Returns:
            int: index of the line if a valid line has been found otherwise returns -1.
        """
        valid_line, _ = nfo.table_line_reader(keyword_store={}, headers=headers_original, line=line)
        if valid_line and len(additional_headers) > 0:
            additional_na_values = ['NA'] * len(additional_headers)
            additional_column_string = ' '.join(additional_na_values)
            split_comments = file_content[index].split('!', 1)
            if len(split_comments) == 1:
                new_completion_line = split_comments[0].replace('\n', '', 1) + ' ' + additional_column_string + '\n'
            else:
                new_completion_line = split_comments[0] + additional_column_string + ' !' + split_comments[1]

            nexusfile_to_write_to, index_in_file = file.find_which_include_file(index)
            if nexusfile_to_write_to.file_content_as_list is None:
                raise ValueError(f'No file content to write to in file: {nexusfile_to_write_to}')
            nexusfile_to_write_to.file_content_as_list[index_in_file] = new_completion_line
        if valid_line:
            return index
        else:
            return -1

    def write_out_new_table_containing_object(self, obj_date: str,
                                              object_properties: dict[str, None | str | float | int],
                                              date_found: bool, new_obj: Any) -> tuple[list[str], int]:
        """Writes out the existing table for an object being added at a new time stamp.

        Args:
            obj_date (str): date for the object to be added at.
            object_properties (dict[str, None | str | float | int]): dictionary containing new attributes of the object.
            date_found (bool): Whether a new TIME card is required or not.
            new_obj (Any): The object getting added that requires the new table to represent it.

        Returns:
            list[str], int: first return arg is a new table containing the requested object and \
            the second representing the line location of the new object within the table.
        """
        nexus_mapping = new_obj.get_keyword_mapping()

        new_table_as_list = ['']
        if not date_found:
            new_table_as_list.append('TIME ' + obj_date)
        new_table_as_list += [self.table_header]
        headers = [k for k, v in nexus_mapping.items() if v[0] in object_properties]
        write_out_headers = ' '.join(headers)
        new_table_as_list.append(write_out_headers)
        new_table_as_list.append(new_obj.to_table_line(headers))
        new_table_as_list.append(self.table_footer)
        new_table_as_list = [x + '\n' if not x.endswith('\n') else x for x in new_table_as_list]
        new_table_as_list.append('\n')
        new_obj_index = len(new_table_as_list) - 1
        return new_table_as_list, new_obj_index

    @staticmethod
    def check_name_date(object_properties: dict[str, None | str | float | int]) -> tuple[str, str]:
        """Checks for the presence of a name and a date in the additional object properties provided."""
        name = object_properties.get('name', None)
        if name is None:
            raise AttributeError(
                'Adding an object requires a name, please provide a "name" entry in the input dictionary.')
        date = object_properties.get('date', None)
        if date is None:
            raise AttributeError(
                'Adding an object requires a date, please provide a "date" entry in the input dictionary.')
        name = str(name)
        date = str(date)

        return name, date

    def add_object_to_file(self, date: str, file_as_list: list[str], file_to_add_to: File, new_object: T,
                           object_properties: dict[str, None | str | float | int]) -> None:
        """Finds where the object should be added based on the date and existing tables.

        Args:
            date (str): date to add the node at.
            file_as_list (list[str]): flattened file as a list of strings.
            file_to_add_to (NexusFile): file to add the new text to.
            new_object (Any): an object with a to_dict, table_header, table_footer, get_keyword_mapping and to_string
            methods.
            object_properties (dict[str, None | str | float | int]): dictionary containing new attributes of the object.

        """
        # initialise some useful variables
        additional_content: list[str] = []
        date_comparison: int = -1
        date_index: int = -1
        insert_line_index: int = -1
        id_line_locs: list[int] = []
        headers: list[str] = []
        additional_headers: list[str] = []
        header_index: int = -1
        last_valid_line_index: int = -1
        headers_original: list[str] = []
        date_found = False
        nexus_mapping = new_object.get_keyword_mapping()
        for index, line in enumerate(file_as_list):
            # check for the dates
            if nfo.check_token('TIME', line):
                date_found_in_file = nfo.get_expected_token_value('TIME', line, [line])
                date_comparison = self.__model._sim_controls.compare_dates(
                    date_found_in_file, date)
                if date_comparison == 0:
                    date_index = index
                    date_found = True
                    continue

            # find a table that exists in that date
            if nfo.check_token(self.table_header, line) and date_index != -1:
                # get the header of the table
                header_index, headers, headers_original = self.get_and_write_new_header(
                    additional_headers, object_properties, file_as_list, index, nexus_mapping, file_to_add_to
                    )
                continue

            if header_index != -1 and index > header_index:
                # check for valid rows + fill extra columns with NA
                line_valid_index = self.fill_in_nas(additional_headers, headers_original, index,
                                                    line, file_to_add_to, file_as_list)
                # set the line to insert the new completion at to be the one after the last valid line
                last_valid_line_index = line_valid_index if line_valid_index > 0 else last_valid_line_index

            # if we've found an existing table then just insert the new object
            if nfo.check_token(self.table_footer, line) and date_comparison == 0:
                insert_line_index = index
                additional_content.append(new_object.to_table_line(headers=headers))
                id_line_locs = [insert_line_index]

            # if we have passed the date or if we're at the end of the file write out the table
            if date_comparison > 0:
                new_table, obj_in_table_index = self.write_out_new_table_containing_object(
                    obj_date=date, object_properties=object_properties, date_found=date_found, new_obj=new_object)
                additional_content.extend(new_table)
                insert_line_index = index
                id_line_locs = [obj_in_table_index + index - 1]

            if insert_line_index >= 0:
                break
        else:
            # if we've finished the loop normally that means we haven't added any additional objects or lines
            # This means we have to add the date and a new table to the end of the file.
            new_table, obj_in_table_index = self.write_out_new_table_containing_object(
                obj_date=date, object_properties=object_properties, date_found=date_found, new_obj=new_object)
            additional_content.extend(new_table)
            insert_line_index = len(file_as_list)
            id_line_locs = [obj_in_table_index + insert_line_index - 1]
        if len(additional_content) == 0:
            raise ValueError('Could not find place to add the additional table lines.')
        # write out to the file_content_as_list
        new_object_ids = {
            new_object.id: id_line_locs
            }
        file_to_add_to.add_to_file_as_list(additional_content=additional_content, index=insert_line_index,
                                           additional_objects=new_object_ids)

    def add_network_obj(self, node_to_add: dict[str, None | str | float | int], obj_type: type[T],
                        network: NexusNetwork) -> T:
        network.get_load_status()
        file_to_add_to = network.get_network_file()
        if self.obj_type is not None:
            obj_type = self.obj_type
        name, date = self.check_name_date(node_to_add)
        new_object = obj_type(node_to_add)
        file_as_list = file_to_add_to.get_flat_list_str_file
        if file_as_list is None:
            raise ValueError(f'No file content found in the surface file specified at {file_to_add_to.location}')
        self.add_object_to_file(date, file_as_list, file_to_add_to, new_object, node_to_add)
        return new_object



"""A class for collecting all NexusWell objects in a NexusSimulator. Handles adding and removing completions
as well as rescheduling wells.
"""
from __future__ import annotations
import warnings
from dataclasses import dataclass, field
from functools import cmp_to_key
from typing import Sequence, Optional, TYPE_CHECKING
from uuid import UUID

import pandas as pd

from ResSimpy.Enums.HowEnum import OperationEnum
from ResSimpy.Nexus.DataModels.NexusCompletion import NexusCompletion
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusWell import NexusWell
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Nexus.NexusKeywords.wells_keywords import WELLS_KEYWORDS
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.Wells import Wells
from ResSimpy.Nexus.load_wells import load_wells
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Utils.invert_nexus_map import attribute_name_to_nexus_keyword

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


@dataclass(kw_only=True)
class NexusWells(Wells):
    """A class for collecting all NexusWell objects in a NexusSimulator. Handles adding and removing completions
    as well as rescheduling wells. This is usually accessed through the model.wells property.

    Arguments:
        model (Simulator): NexusSimulator object that has the instance of wells on.
    """
    __model: NexusSimulator
    _wells: list[NexusWell] = field(default_factory=list)
    __date_format: DateFormat

    def __init__(self, model: NexusSimulator) -> None:
        self.__model = model
        self.__add_object_operations = AddObjectOperations(NexusCompletion, self.table_header, self.table_footer, model)
        super().__init__()

    @property
    def date_format(self) -> DateFormat:
        if not self._wells_loaded:
            self._load()
        return self.__date_format

    @property
    def table_header(self) -> str:
        return 'WELLSPEC'

    @property
    def table_footer(self) -> str:
        return ''

    def get_all(self) -> Sequence[NexusWell]:
        if not self._wells_loaded:
            self._load()
        return self._wells

    def get(self, well_name: str) -> Optional[NexusWell]:
        """Returns a specific well requested, or None if that well cannot be found."""
        if not self._wells_loaded:
            self._load()

        wells_to_return = filter(lambda x: x.well_name.upper() == well_name.upper(), self._wells)

        return next(wells_to_return, None)

    def get_df(self) -> pd.DataFrame:
        # loop through wells and completions to output a table
        if not self._wells_loaded:
            self._load()

        store_dictionaries = []
        for well in self._wells:
            for completion in well.completions:
                completion_props: dict[str, None | float | int | str] = {
                    'well_name': well.well_name,
                    'units': well.unit_system.name,
                    }
                completion_props.update(completion.to_dict())
                store_dictionaries.append(completion_props)
        df_store = pd.DataFrame(store_dictionaries)
        df_store = df_store.dropna(axis=1, how='all')
        return df_store

    def _load(self) -> None:
        if self.__model.model_files.well_files is None:
            raise FileNotFoundError('No wells files found for current model.')
        for method_number, well_file in self.__model.model_files.well_files.items():
            if well_file.location is None:
                warnings.warn(f'Well file location has not been found for {well_file}')
                continue
            new_wells, date_format = load_wells(nexus_file=well_file, start_date=self.__model.start_date,
                                                default_units=self.__model.default_units,
                                                model_date_format=self.__model.date_format)

            self._wells += new_wells
            self.__date_format = date_format
        self._wells_loaded = True

    def modify(self, well_name: str, completion_properties_list: list[dict[str, None | float | int | str]],
               how: OperationEnum = OperationEnum.ADD) -> None:
        """Modify the existing wells in memory using a dictionary of properties.

        Args:
        ----
            well_name (str): name of the well to modify
            completion_properties_list (list[InputDict]): a dictionary containing the properties to modify with the \
                attribute as keys and the values as the updated property value. If remove will remove perforation that \
                matches the values in the dictionary.
            how (OperationEnum): operation enum taking the values OperationEnum.ADD, OperationEnum.REMOVE. \
                Specifies how to modify the existing wells perforations.
            remove_all_that_match (bool): If True will remove all wells that partially match the completion_properties\
                provided. If False will remove perforation if only one matches, if several match throws a warning and \
                does not remove them.
            write_to_file (bool): If True writes directly to file. (Currently not in use)
        """
        well = self.get(well_name)
        if well is None:
            raise ValueError(f'No well named {well_name} found in simulator')
        for perf in completion_properties_list:
            if how == OperationEnum.ADD:
                try:
                    date = perf.get('date')
                except AttributeError:
                    raise AttributeError(
                        f'No date provided in perf: {perf}, please provide a date to add the perforation at.')
                if date is None:
                    raise AttributeError(
                        f'No date provided in perf: {perf}, please provide a date to add the perforation at.')
                self.add_completion(well_name=well_name, completion_properties=perf)
            elif how == OperationEnum.REMOVE:
                completions_to_remove = well.find_completions(perf)
                well._remove_completions_from_memory(completions_to_remove)
            elif how == OperationEnum.MODIFY:
                raise NotImplementedError('Modify in place not yet available. Please choose one of ADD/REMOVE')
            else:
                raise ValueError('Please select one of the valid OperationEnum values: e.g. OperationEnum.ADD')

    def add_completion(self, well_name: str, completion_properties: dict[str, None | float | int | str],
                       preserve_previous_completions: bool = True, comments: Optional[str] = None) -> None:
        """Adds a completion to an existing wellspec file.

        Args:
        ----
            well_name (str): well name to update
            completion_properties (dict[str, float | int | str): properties of the completion you want to update.
            Must contain date of the completion to be added.
            preserve_previous_completions (bool): if true a new perforation added on a TIME card without a \
            wellspec card for that well will preserve the previous completions from the closest TIME card in addition \
            to the new completion
        """
        _, completion_date = self.__add_object_operations.check_name_date({'name': well_name} | completion_properties)
        well = self.get(well_name)
        if well is None:
            # TODO could make this not raise an error and instead initialize a NexusWell and add it to NexusWells
            raise ValueError(
                f"No well found named: {well_name}. Cannot add completion to a well that doesn't exist")
        well_id = well.completions[0].id

        # add completion in memory
        new_completion = well._add_completion_to_memory(date=completion_date,
                                                        completion_properties=completion_properties,
                                                        date_format=self.date_format)

        if self.__model.model_files.well_files is None:
            raise FileNotFoundError('No well file found, cannot modify ')

        wellspec_file = self.__add_object_operations.find_which_file_from_id(obj_id=well_id,
                                                                             file_type_to_search='well_files')

        # initialise some storage variables
        nexus_mapping = NexusCompletion.get_keyword_mapping()
        new_completion_time_index = -1
        header_index = -1
        headers: list[str] = []
        headers_original: list[str] = []
        additional_headers: list[str] = []
        file_content = wellspec_file.get_flat_list_str_file
        date_found = False
        new_completion_index = len(file_content)
        new_completion_string: list[str] = []
        last_valid_line_index: int = -1
        writing_new_wellspec_table = False
        date_comp = 0
        # if no time cards present in the file just find the name of the well instead
        if not nfo.value_in_file('TIME', file_content):
            new_completion_time_index = 0

        for index, line in enumerate(file_content):
            if header_index == -1 and nfo.check_token('TIME', line):
                wellspec_date = nfo.get_expected_token_value('TIME', line, [line])
                date_comp = self.__model._sim_controls.compare_dates(wellspec_date, completion_date)
                if date_comp == 0:
                    # if we've found the date we're looking for start looking for a wellspec and name card
                    new_completion_time_index = index
                    date_found = True
                    continue
                elif date_comp > 0 and header_index == -1:
                    # if no date is found that is equal and we've found a date that is greater than the specified date
                    # start to compile a wellspec table from scratch and add in the time cards
                    new_completion_index = index
                    header_index = index - 1
                    headers, new_completion_index, new_completion_string, found_completion_at_previous_date = \
                        self.__write_out_existing_wellspec(
                            completion_date, completion_properties, date_found, index,
                            new_completion_index, preserve_previous_completions, well, well_name)
                    writing_new_wellspec_table = True
                    if not found_completion_at_previous_date:
                        break
                else:
                    continue

            if nfo.check_token('WELLSPEC', line) and nfo.get_token_value('WELLSPEC', line, [line]) == well_name \
                    and new_completion_time_index != -1:
                # get the header of the wellspec table
                header_index, headers, headers_original = self.__add_object_operations.get_and_write_new_header(
                    additional_headers, completion_properties, file_content, index, nexus_mapping, wellspec_file
                    )
                continue

            if header_index != -1 and nfo.nexus_token_found(line, WELLS_KEYWORDS) and index > header_index:
                # if we hit the end of the wellspec table for the given well set the index for the new completion
                if last_valid_line_index != -1:
                    new_completion_index = last_valid_line_index + 1
                break

            elif header_index != -1 and index > header_index:
                # check for valid rows + fill extra columns with NA
                line_valid_index = self.__add_object_operations.fill_in_nas(additional_headers, headers_original, index,
                                                                            line, wellspec_file, file_content)
                # set the line to insert the new completion at to be the one after the last valid line
                last_valid_line_index = line_valid_index if line_valid_index > 0 else last_valid_line_index
        # If we haven't found a TIME card after the for loop then we haven't got a valid date so add it at the end
        if date_comp < 0:
            new_completion_index = len(file_content)
            headers, new_completion_index, new_completion_string, found_completion_at_previous_date = \
                self.__write_out_existing_wellspec(
                    completion_date, completion_properties, date_found, new_completion_index,
                    new_completion_index, preserve_previous_completions, well, well_name)
            writing_new_wellspec_table = True

        # construct the new completion and ensure the order of the values is in the same order as the headers
        new_completion_string += [new_completion.to_table_line(headers)]
        new_completion_additional_lines = len(new_completion_string)
        if writing_new_wellspec_table:
            new_completion_string += ['\n']
        # write out to the file_content_as_list
        new_completion_object_ids = {new_completion.id: [new_completion_index + new_completion_additional_lines - 1]}
        wellspec_file.add_to_file_as_list(additional_content=new_completion_string, index=new_completion_index,
                                          additional_objects=new_completion_object_ids, comments=comments)

    def __write_out_existing_wellspec(self, completion_date: str,
                                      completion_properties: dict[str, None | float | int | str],
                                      date_found: bool, index: int, new_completion_index: int,
                                      preserve_previous_completions: bool, well: NexusWell, well_name: str) -> \
            tuple[list[str], int, list[str], bool]:
        """Writes out the existing wellspec for a well at a new time stamp."""
        nexus_mapping = NexusCompletion.get_keyword_mapping()
        completion_table_as_list = ['\n']
        if not date_found:
            completion_table_as_list += ['TIME ' + completion_date + '\n']
        completion_table_as_list += ['WELLSPEC ' + well_name + '\n']
        headers = [k for k, v in nexus_mapping.items() if v[0] in completion_properties]
        if preserve_previous_completions:
            # get all the dates for that well
            date_list = well.dates_of_completions
            previous_dates = [x for x in date_list if
                              self.__model._sim_controls.compare_dates(x, completion_date) < 0]
            if len(previous_dates) == 0:
                # if no dates that are smaller than the completion date then only add the perforation
                # at the current index with a new wellspec card.
                warnings.warn(f'No previous completions found for {well_name} at date: {completion_date}')
                new_completion_index = index
                write_out_headers = [' '.join(headers) + '\n']
                completion_table_as_list += write_out_headers
                return headers, new_completion_index, completion_table_as_list, False

            # get the most recent date that is earlier than the new completion date
            previous_dates = sorted(previous_dates, key=cmp_to_key(self.__model._sim_controls.compare_dates))
            last_date = str(previous_dates[-1])
            completion_to_find: dict[str, None | float | int | str] = {'date': last_date}
            # find all completions at this date
            previous_completion_list = well.find_completions(completion_to_find)
            if len(previous_completion_list) > 0:
                prev_completion_properties = {k: v for k, v in previous_completion_list[0].to_dict().items() if
                                              v is not None}
                for key in prev_completion_properties:
                    if key == 'date':
                        continue
                    header_key = attribute_name_to_nexus_keyword(nexus_mapping, key)
                    if header_key not in headers:
                        headers.append(header_key)

            write_out_headers = [' '.join(headers) + '\n']
            completion_table_as_list += write_out_headers
            # run through the existing completions to duplicate the completion at the new time
            for completion in previous_completion_list:
                completion_table_as_list += [completion.to_table_line(headers)]
        else:
            write_out_headers = [' '.join(headers) + '\n']
            completion_table_as_list += write_out_headers
        return headers, new_completion_index, completion_table_as_list, True

    def remove_completion(self, well_name: str,
                          completion_properties: Optional[dict[str, None | float | int | str]] = None,
                          completion_id: Optional[UUID] = None) -> None:

        well = self.get(well_name)
        if well is None:
            raise ValueError(f'No well found with name: {well_name}')

        if completion_properties is None and completion_id is None:
            raise ValueError('Must provide one of completion_properties dictionary or completion_id.')

        # check for a date:
        if completion_properties is not None:
            completion_date = completion_properties.get('date', 'NO_DATE_PROVIDED')
            if completion_date == 'NO_DATE_PROVIDED':
                raise AttributeError('Completion requires a date. '
                                     'Please provide a date in the completion_properties_list dictionary.')
            if completion_id is None:
                completion_id = well.find_completion(completion_properties).id
        if completion_id is None:
            raise ValueError('No completion found for completion_properties')
        # find which wellspec file we should edit
        wellspec_file = self.__add_object_operations.find_which_file_from_id(obj_id=completion_id,
                                                                             file_type_to_search='well_files')

        # remove from the well object/wells class
        completion_date = well.get_completion_by_id(completion_id).date
        well._remove_completion_from_memory(completion_to_remove=completion_id)

        # drop it from the wellspec file or include file if stored in include file
        if wellspec_file.object_locations is None:
            raise ValueError(f'No object locations specified, cannot find completion id: {completion_id}')
        completion_indices = wellspec_file.object_locations[completion_id]
        if len(completion_indices) > 0:
            for comp_index in completion_indices:
                wellspec_file.remove_from_file_as_list(comp_index, [completion_id])

        # check that we have completions left:
        find_completions_dict: dict[str, None | float | int | str] = {'date': completion_date}
        remaining_completions = well.find_completions(find_completions_dict)
        if len(remaining_completions) == 0:
            # if there are no more completions remaining for that time stamp then remove the wellspec header!
            self.__remove_wellspec_header(str(completion_date), well_name, wellspec_file)

    def __remove_wellspec_header(self, completion_date: str, well_name: str, wellspec_file: NexusFile) -> None:
        """Removes the wellspec and header if the wellspec table is empty\
        must first check for whether the well has any remaining completions in the wellspec table.
        """
        nexus_mapping = NexusCompletion.get_keyword_mapping()
        completion_date_found = False
        file_content = wellspec_file.get_flat_list_str_file
        wellspec_index = -1
        header_index = -1
        for index, line in enumerate(file_content):
            if nfo.check_token('TIME', line) and nfo.get_expected_token_value('TIME', line, [line]) == \
                    completion_date:
                completion_date_found = True
            if completion_date_found and nfo.check_token('WELLSPEC', line) and \
                    nfo.get_token_value('WELLSPEC', line, [line]) == well_name:
                # get the index in the list as string
                wellspec_index = index
                keyword_map = {x: y[0] for x, y in nexus_mapping.items()}
                wellspec_table = file_content[wellspec_index::]
                header_index, _ = nfo.get_table_header(file_as_list=wellspec_table, header_values=keyword_map)
                header_index += wellspec_index
                break
        wellspec_file.remove_from_file_as_list(header_index)
        wellspec_file.remove_from_file_as_list(wellspec_index)

    def modify_completion(self, well_name: str, properties_to_modify: dict[str, None | float | int | str],
                          completion_to_change: Optional[dict[str, None | float | int | str]] = None,
                          completion_id: Optional[UUID] = None,
                          comments: Optional[str] = None) -> None:
        """Modify an existing matching completion, preserves attributes and modifies only additional properties
        found within the provided properties to modify dictionary.

        Args:
            well_name (str): Name of the well with the completion to be modified.
            properties_to_modify (dict[str, None | float | int | str]): attributes to change to.
            completion_to_change (Optional[dict[str, None | float | int | str]]): properties of the existing completion.
            User must provide enough to uniquely identify the completion.
            completion_id (Optional[UUID]): If provided will match against a known UUID for the completion.
        """
        well = self.get(well_name)
        if well is None:
            raise ValueError(f'No well found with name: {well_name}')

        if completion_to_change is not None:
            completion = well.find_completion(completion_to_change)
            completion_id = completion.id
        elif completion_id is not None:
            completion = well.get_completion_by_id(completion_id)
        else:
            raise ValueError('Must provide one of completion_to_change dictionary or completion_id')

        # start with the existing properties
        update_completion_properties: dict[str, None | float | int | str] = \
            {k: v for k, v in completion.to_dict().items() if v is not None}

        update_completion_properties.update(properties_to_modify)

        self.remove_completion(well_name, completion_id=completion_id)
        self.add_completion(well_name, update_completion_properties, preserve_previous_completions=True,
                            comments=comments)



from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from ResSimpy.Grid import VariableEntry
import ResSimpy.Nexus.nexus_file_operations as nfo

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


class StructuredGridOperations:
    def __init__(self, model: NexusSimulator) -> None:
        self.__model: NexusSimulator = model

    @staticmethod
    def load_token_value_if_present(token: str, modifier: str, token_property: VariableEntry,
                                    line: str, file_as_list: list[str],
                                    ignore_values: Optional[list[str]] = None) -> None:
        """Gets a token's value if there is one and loads it into the token_property.

        Args:
        ----
            token (str): the token being searched for. e.g. 'PERMX'
            modifier (str): any modifiers applied to the token e.g. 'MULT'
            token_property (VariableEntry): VariableEntry object to store the modifier and value pair into
            line (str): line to search for the token in
            ignore_values (Optional[list[str]], optional): values to be ignored. Defaults to None.

        Raises:
        ------
            ValueError: raises an error if no numerical value can be found after the supplied token modifier pair
        Returns:
            None: stores the value into token_property supplied instead
        """

        if ignore_values is None:
            ignore_values = []
        token_modifier = f"{token} {modifier}"

        if nfo.check_token(token, line) and nfo.get_token_value(token, line, file_as_list) == modifier:
            # If we are loading a multiple, load the two relevant values, otherwise just the next value
            if modifier == 'MULT':
                numerical_value = nfo.get_expected_token_value(token_modifier, line, file_as_list, ignore_values=None)
                if numerical_value is None:
                    raise ValueError(
                        f'No numerical value found after {token_modifier} keyword in line: {line}')
                value_to_multiply = nfo.get_token_value(token_modifier, line, file_as_list,
                                                        ignore_values=[numerical_value])
                if numerical_value is not None and value_to_multiply is not None:
                    token_property.modifier = 'MULT'
                    token_property.value = f"{numerical_value} {value_to_multiply}"
            else:
                value = nfo.get_token_value(token_modifier, line, file_as_list, ignore_values=ignore_values)
                if value is None:
                    # Could be 'cut short' by us excluding the rest of a file
                    token_property.value = None
                    token_property.modifier = modifier
                token_property.value = value
                token_property.modifier = modifier

    @staticmethod
    def replace_value(file_as_list: list[str], old_property: VariableEntry, new_property: VariableEntry,
                      token_name: str) -> None:
        """Replace the value and token + modifier with the new values.

        Args:
        ----
            file_as_list (list[str]): a list of strings containing each line of the file as a new entry
            old_property (VariableEntry): property found in the original file to be replaced
            new_property (VariableEntry): new property to replace the old property with
            token_name (str): name of the token being replaced
        Returns:
            None: modifies the file_as_list with the new property
        """

        for line in file_as_list:
            old_token_modifier = f"{token_name} {old_property.modifier}"
            new_token_modifier = f"{token_name} {new_property.modifier}"
            ignore_values = ['INCLUDE'] if old_property.modifier == 'VALUE' else []
            if nfo.check_token(old_token_modifier, line):
                # If we are replacing a mult, replace the first value with a blank
                if old_property.modifier == 'MULT':
                    dummy_value = VariableEntry('MULT', '')
                    nfo.get_token_value(old_token_modifier, line, file_as_list, ignore_values=ignore_values,
                                        replace_with=dummy_value)

                nfo.get_token_value(old_token_modifier, line, file_as_list, ignore_values=ignore_values,
                                    replace_with=new_property)

                new_line = line.replace(old_token_modifier, new_token_modifier, 1)
                line_index = file_as_list.index(line)
                file_as_list[line_index] = new_line

    @staticmethod
    def append_include_to_grid_file(include_file_location: str, structured_grid_file_path: str) -> None:
        # TODO: change append to be an optional parameter
        """Appends an include file to the end of a grid for adding LGRs.

        Args:
        ----
            include_file_location (str): path to a file to include in the grid
            structured_grid_file_path (str): file path to the structured grid.

        Raises:
        ------
            ValueError: if no structured grid file path is specified in the class instance
        """
        # Get the existing file as a list
        if structured_grid_file_path is None:
            raise ValueError("No file path given or found for structured grid file path. \
                Please update structured grid file path")
        file = nfo.load_file_as_list(structured_grid_file_path)

        file.append(f"\nINCLUDE {include_file_location}\n")
        file.append("TOLPV LGR1 0\n")

        # Save the new file contents
        new_file_str = "".join(file)
        with open(structured_grid_file_path, "w") as text_file:
            text_file.write(new_file_str)

    @staticmethod
    def get_grid_file_as_3d_list(path: str) -> Optional[list]:
        """Converts a grid file to a 3D list.

        Args:
        ----
            path (str): path to a grid file

        Returns:
        -------
            Optional[list[str]]: Returns None if no file is found, returns the grid as a 3d array otherwise
        """
        try:
            with open(path) as f:
                grid_file_list = list(f)
        except FileNotFoundError:
            return None

        sub_lists = []

        new_list_str = ""
        for sub_list in grid_file_list[4:]:
            if sub_list == '\n':
                new_list_split = [x.split("\t") for x in new_list_str.split("\n")]
                new_list_split_cleaned = []
                for x_list in new_list_split:
                    float_list_split = [float(x) for x in x_list if x != ""]
                    new_list_split_cleaned.append(float_list_split)
                sub_lists.append(new_list_split_cleaned)
                new_list_str = ""
            else:
                new_list_str = new_list_str + sub_list
        return sub_lists

    def view_command(self, field: str, previous_lines: int = 3, following_lines: int = 3) -> Optional[str]:
        """Displays how the property is declared in the structured grid file.

        Args:
        ----
            field (str): property as written in the structured grid (e.g. KX)
            previous_lines (int, optional): how many lines to look back from the field searched for. Defaults to 3.
            following_lines (int, optional): how many lines to look forward from the field searched for. Defaults to 3.

        Raises:
        ------
            ValueError: if no structured grid file path is specified in the class instance

        Returns:
        -------
            Optional[str]: the string associated with the supplied property from within the structured grid. \
                If the field is not found in the structured grid returns None.
        """
        structured_grid_dict = self.__model.get_structured_grid_dict()
        structured_grid_path = self.__model.structured_grid_path
        command_token = f"{field.upper()} {structured_grid_dict[field.lower()].modifier}"
        if structured_grid_path is None:
            raise ValueError("No path found for structured grid file path. \
                Please provide a path to the structured grid")
        file_as_list = nfo.load_file_as_list(structured_grid_path)

        for line in file_as_list:
            if nfo.check_token(command_token, line):
                start_index = file_as_list.index(line) - previous_lines \
                    if file_as_list.index(line) - previous_lines > 0 else 0
                end_index = file_as_list.index(line) + following_lines \
                    if file_as_list.index(line) + following_lines < len(file_as_list) else len(file_as_list) - 1

                new_array = file_as_list[start_index: end_index]
                new_array = [x.strip("'") for x in new_array]
                value = "".join(new_array)
                return value
        return None



"""Class for Nexus relative permeability endpoints data storage."""
from __future__ import annotations
from dataclasses import dataclass
from typing import Optional
from ResSimpy.RelPermEndPoint import RelPermEndPoint


@dataclass
class NexusRelPermEndPoint(RelPermEndPoint):
    """Class for Nexus relative permeability endpoints data storage."""
    def __init__(self, swl: Optional[float] = None, swr: Optional[float] = None, swu: Optional[float] = None,
                 sgl: Optional[float] = None, sgr: Optional[float] = None, sgu: Optional[float] = None,
                 swro: Optional[float] = None, sgro: Optional[float] = None, sgrw: Optional[float] = None,
                 krw_swro: Optional[float] = None, krw_swu: Optional[float] = None, krg_sgro: Optional[float] = None,
                 krg_sgu: Optional[float] = None, kro_swl: Optional[float] = None, kro_swr: Optional[float] = None,
                 kro_sgl: Optional[float] = None, kro_sgr: Optional[float] = None, krw_sgl: Optional[float] = None,
                 krw_sgr: Optional[float] = None, krg_sgrw: Optional[float] = None, sgtr: Optional[float] = None,
                 sotr: Optional[float] = None) -> None:
        super().__init__(swl=swl, swr=swr, swu=swu, sgl=sgl, sgr=sgr, sgu=sgu, swro=swro, sgro=sgro,
                         sgrw=sgrw, krw_swro=krw_swro, krw_swu=krw_swu, krg_sgro=krg_sgro, krg_sgu=krg_sgu,
                         kro_swl=kro_swl, kro_swr=kro_swr, kro_sgl=kro_sgl, kro_sgr=kro_sgr, krw_sgl=krw_sgl,
                         krw_sgr=krw_sgr, krg_sgrw=krg_sgrw, sgtr=sgtr, sotr=sotr)

    @staticmethod
    def nexus_mapping() -> dict[str, tuple[str, type]]:
        """Returns a dictionary of mapping from nexus keyword to attribute name."""
        nexus_mapping: dict[str, tuple[str, type]] = {
            'SWL': ('swl', float),
            'SWR': ('swr', float),
            'SWU': ('swu', float),
            'SGL': ('sgl', float),
            'SGR': ('sgr', float),
            'SGU': ('sgu', float),
            'SWRO': ('swro', float),
            'SGRO': ('sgro', float),
            'SGRW': ('sgrw', float),
            'KRW_SWRO': ('krw_swro', float),
            'KRW_SWU': ('krw_swu', float),
            'KRG_SGRO': ('krg_sgro', float),
            'KRG_SGU': ('krg_sgu', float),
            'KRO_SWL': ('kro_swl', float),
            'KRO_SWR': ('kro_swr', float),
            'KRO_SGL': ('kro_sgl', float),
            'KRO_SGR': ('kro_sgr', float),
            'KRW_SGL': ('krw_sgl', float),
            'KRW_SGR': ('krw_sgr', float),
            'KRG_SGRW': ('krg_sgrw', float),
            'SGTR': ('sgtr', float),
            'SOTR': ('sotr', float),
        }
        return nexus_mapping



"""ResSimpy Nexus Module, containing the Nexus derived classes and Nexus specific modules."""


__all__ = ['load_wells', 'NexusWells', 'NexusSimulator', 'DataModels', 'array_function_operations', 'constants',
           'logfile_operations', 'nexus_file_operations', 'NexusGrids', 'NexusNetwork', 'NexusReporting',
           'rel_perm_operations', 'structured_grid_operations', 'runcontrol_operations']



from __future__ import annotations
from typing import Optional
import ResSimpy.Nexus.nexus_file_operations as nfo


def get_relperm_combined_fluid_column_heading(table_heading: str) -> str:
    """Returns the combined rel perm fluid perm heading for a Nexus Relperm table.

    Args:
        table_heading (str): heading for the rel perm table, expects one of (WOTABLE, GOTABLE, GWTABLE).

    Returns:
        str: one of (KROW, KROG, KRWG) defaults to KRWG if one of the expected table_headings is not given
    """
    rel_perm_header_map = {
        'WOTABLE': 'KROW',
        'GOTABLE': 'KROG',
        'GWTABLE': 'KRWG',
    }
    column_heading = rel_perm_header_map.get(table_heading, 'KRWG')

    return column_heading


def get_relperm_single_fluid_column_heading(table_heading: str) -> str:
    """Returns the single fluid heading for a Nexus Relperm table.

    Args:
        table_heading (str): heading for the rel perm table, expects one of (WOTABLE, GOTABLE, GWTABLE)

    Returns:
        str: heading for the single fluid rel perm header one of (KRG, KRW)
    """
    if table_heading == 'GOTABLE':
        column_heading = 'KRG'
    else:
        column_heading = 'KRW'

    return column_heading


def get_relperm_base_saturation_column_heading(table_heading: str) -> str:
    """Returns the column heading for the base saturation column.

    Args:
        table_heading (str): heading for the rel perm table, expects one of (WOTABLE, GOTABLE, GWTABLE)

    Returns:
        str: heading for the saturation based on the table header one of (SG, SW)
    """

    if table_heading == 'GOTABLE':
        column_heading = 'SG'
    else:
        column_heading = 'SW'

    return column_heading


def load_nexus_relperm_table(relperm_file_path: str) -> dict[str, list[tuple[float, float]]]:
    """Loads in a Nexus relperm table and returns a dictionary with two lists, one with the relperm values for the
    single fluid, and the other with the values for combined fluids.

    Args:
        relperm_file_path (str): path to a single Nexus rel perm file.

    Raises:
        ValueError: if the table header cannot be found for the rel perm table

    Returns:
        dict[str, list[tuple[float, float]]]: dictionary containing two entries one for the column (KRG, KRW)
        and one for one of (KROW, KROG, KRWG) depending on the type of table read.
        Each list entry consists of a tuple of (saturation, relperm value)
    """

    file_as_list = nfo.load_file_as_list(relperm_file_path)

    # Find the column headings line (assume it is the first non-commented out line after the table heading)
    possible_table_headings = ['GWTABLE', 'WOTABLE', 'GOTABLE']
    header_index = None
    table_heading = None

    for index, line in enumerate(file_as_list):
        first_value_in_line = nfo.get_next_value(0, [line], line)
        if first_value_in_line in possible_table_headings:
            table_heading = first_value_in_line
            header_index = index + 1
            break

    if header_index is None or table_heading is None:
        raise ValueError("Cannot find the header for this relperm table")

    # Read in the header line to get the column order
    header_line = file_as_list[header_index]
    columns: list[str] = []

    next_column_heading = nfo.get_next_value(0, [header_line], header_line)
    next_line = header_line
    while next_column_heading is not None:
        columns.append(next_column_heading)
        next_line = next_line.replace(next_column_heading, "", 1)
        next_column_heading = nfo.get_next_value(0, [next_line], next_line)

    # Load in each row from the table
    all_values: list[Optional[dict[str, str]]] = []

    for line in file_as_list[header_index + 1:]:
        trimmed_line = line
        line_values: Optional[dict[str, str]] = {}
        for column in columns:
            value = nfo.get_next_value(0, [trimmed_line], trimmed_line)

            # If we hit a comment or blank line, assume that we've reached the end of our table
            if value is None:
                line_values = None
                break

            trimmed_line = trimmed_line.replace(value, "", 1)

            if line_values is not None:
                line_values[column] = value
        if line_values is not None:
            all_values.append(line_values)
        elif len(all_values) > 0:
            # We've reached the end of the table, finish reading the lines now
            break

    # Retrieve the water and gas values, and return them
    single_fluid_relperms = []  # E.g. Water
    combined_fluid_relperms = []  # E.g. Water-Oil

    single_fluid_column_heading = get_relperm_single_fluid_column_heading(
        table_heading)
    combined_fluid_column_heading = get_relperm_combined_fluid_column_heading(
        table_heading)
    base_saturation_heading = get_relperm_base_saturation_column_heading(
        table_heading)

    for index, row in enumerate(all_values):
        if row is None:
            continue
        single_fluid_relperms.append(
            (float(row[base_saturation_heading]),
             float(row[single_fluid_column_heading]))
        )
        combined_fluid_relperms.append(
            (float(row[base_saturation_heading]), float(
                row[combined_fluid_column_heading]))
        )

    return {'single_fluid': single_fluid_relperms, 'combined_fluids': combined_fluid_relperms}



def invert_nexus_map(nexus_mapping: dict[str, tuple[str, type]]) -> dict[str, str]:
    r"""Takes a nexus map of the form {NEXUS KEYWORD: ('attribute_name', type), ...} and reverses it to
    give a map from attribute names to Nexus Keywords.

    Args:
    ----
        nexus_mapping (dict[str, tuple[str, type]]): dictionary of the form \n
        {NEXUS KEYWORD: ('attribute_name', type), ...}

    Returns:
    -------
        dict[str, str] of the form {'attribute_name': 'NEXUS_KEYWORD'}

    """
    inverted_nexus_map = {v[0]: k for k, v in nexus_mapping.items()}
    return inverted_nexus_map


def nexus_keyword_to_attribute_name(nexus_mapping: dict[str, tuple[str, type]], nexus_keyword: str) -> str:
    """Takes a Nexus Keyword and maps to the corresponding attribute name from a given nexus mapping."""
    attribute_name_tuple = nexus_mapping.get(nexus_keyword.upper(), None)
    if attribute_name_tuple is None:
        raise AttributeError(f'No nexus keyword found named "{nexus_keyword.upper()}" in the provided nexus mapping')
    return attribute_name_tuple[0]


def attribute_name_to_nexus_keyword(nexus_mapping: dict[str, tuple[str, type]], attribute_name: str) -> str:
    """Takes an attribute name and maps to the corresponding nexus keyword from a given nexus mapping."""

    invert_map = invert_nexus_map(nexus_mapping)
    nexus_keyword = invert_map.get(attribute_name.lower(), None)
    if nexus_keyword is None:
        raise AttributeError(f'No attribute found with name {attribute_name.lower()} in the provided nexus mapping')
    return nexus_keyword
